/*
 * Copyright (c) 2016 BiTel Inc. All rights reserved.
*/
package ru.bitel.mybgbilling.kernel.common;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Serializable;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.text.DateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import javax.activation.DataHandler;
import javax.annotation.PostConstruct;
import javax.enterprise.context.SessionScoped;
import javax.faces.application.FacesMessage;
import javax.faces.component.UIComponent;
import javax.faces.component.UIViewRoot;
import javax.faces.context.ExternalContext;
import javax.faces.context.FacesContext;
import javax.inject.Inject;
import javax.inject.Named;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.Part;
import javax.xml.ws.soap.SOAPFaultException;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import ru.bitel.bgbilling.common.BGException;
import ru.bitel.bgbilling.common.BGMessageException;
import ru.bitel.bgbilling.common.BGRuntimeException;
import ru.bitel.common.Utils;
import ru.bitel.common.function.ThrowingSupplier;
import ru.bitel.common.io.IOUtils;
import ru.bitel.common.util.FileDownload;
import ru.bitel.mybgbilling.kernel.common.utils.CommonUtils;
import ru.bitel.mybgbilling.kernel.navigation.LocaleBean;
import ru.bitel.mybgbilling.kernel.navigation.NavigationBean;
import ru.bitel.mybgbilling.kernel.ws.CustomerBean;

public class AbstractBean
	implements Serializable
{
	private static final long serialVersionUID = 1L;

	private static final Logger logger = LogManager.getLogger();

	@Inject
	protected CustomerBean customerBean;

	@Inject
	protected NavigationBean navigationBean;

	@Inject
	protected LocaleBean localeBean;

	private AtomicInteger contractId;
	private int moduleId;

	protected int step = 0;

	private final String beanName;
	private final Method populateMethod;

	public AbstractBean()
	{
		final Class<?> clazz = this.getClass();

		if( clazz.getAnnotation( SessionScoped.class ) != null )
		{
			String name;

			final Named named = clazz.getAnnotation( Named.class );
			if( named != null && Utils.notBlankString( named.value() ) )
			{
				name = named.value();
			}
			else
			{
				name = clazz.getSimpleName();

				int idx = name.indexOf( '$' );
				if( idx > 0 )
				{
					name = name.substring( 0, 1 ).toLowerCase() + name.substring( 1, idx );
				}
				else
				{
					name = name.substring( 0, 1 ).toLowerCase() + name.substring( 1 );
				}
			}

			this.beanName = name;

			Method populateMethod = null;

			try
			{
				populateMethod = clazz.getMethod( "populate" );
			}
			catch( Exception ex )
			{}

			this.populateMethod = populateMethod;
		}
		else
		{
			this.beanName = null;
			this.populateMethod = null;
		}
	}

	private long lastAutoPopulateMillis = -1L;

	public void autoPopulate()
		throws BGException
	{
		if( populateMethod != null && beanName != null )
		{
			if( System.currentTimeMillis() - lastAutoPopulateMillis < TimeUnit.SECONDS.toMillis( 30 ) )
			{
				return;
			}

			UIViewRoot viewRoot = FacesContext.getCurrentInstance().getViewRoot();
			UIComponent component = viewRoot.findComponent( beanName + "-populate" );

			if( component != null )
			{
				try
				{
					populateMethod.invoke( this );
				}
				catch( IllegalAccessException | IllegalArgumentException | InvocationTargetException e )
				{
					logger.error( e.getMessage(), e );
				}

				lastAutoPopulateMillis = System.currentTimeMillis();
			}
		}
	}

	public int getStep()
	{
		return step;
	}

	public void setStep( int step )
	{
		this.step = step;
	}

	@PostConstruct
	protected void postConstuct()
	{
		try
		{
			this.contractId = customerBean.getContractIdRef();
			this.moduleId = navigationBean.getModuleId();

			init();
		}
		catch( BGException e )
		{
			logger.error( e.getMessage(), e );
		}
	}

	protected void init()
		throws BGException
	{

	}

	protected void processException( Logger logger, String clientId, BGRuntimeException ex )
	{
		if( ex.getCause() instanceof BGMessageException )
		{
			BGMessageException me = (BGMessageException)ex.getCause();
			ru.bitel.mybgbilling.kernel.common.utils.Utils.addTagErrorMessage( FacesContext.getCurrentInstance(), clientId, ex.getTag(), me.getMessage() );

			if( logger == null )
			{
				logger = AbstractBean.logger;
			}

			logger.info( String.format( "BGMessageException: [%s] %s", me.getTag(), me.getMessage() ) );
		}
		else
		{
			throw ex;
		}
	}

	protected void addErrorMessage( String clientId, String message )
	{
		ru.bitel.mybgbilling.kernel.common.utils.Utils.addErrorMessage( FacesContext.getCurrentInstance(), clientId, message );
		//FacesContext.getCurrentInstance().addMessage( clientId, new FacesMessage( FacesMessage.SEVERITY_ERROR, "Error Message", message ) );
	}

	protected void addInfoMessage( String clientId, String message )
	{
		FacesContext.getCurrentInstance().addMessage( clientId, new FacesMessage( FacesMessage.SEVERITY_INFO, "Info Message", message ) );
	}

	protected List<String> dateListToStringList( List<Date> dateList )
	{
		List<String> result = new ArrayList<String>();

		if( dateList == null )
		{
			return result;
		}

		Locale locale = FacesContext.getCurrentInstance().getViewRoot().getLocale();
		if( locale == null )
		{
			locale = Locale.getDefault();
		}

		DateFormat format = DateFormat.getDateInstance( DateFormat.MEDIUM, locale );

		for( Date date : dateList )
		{
			result.add( format.format( date ) );
		}

		return result;
	}

	protected <E> List<E> reverse( List<E> list )
	{
		if( list == null )
		{
			return Utils.maskNull( list );
		}

		Collections.reverse( list );
		return list;
	}

	protected int getContractId()
	{
		return contractId.get();
	}

	public int getModuleId()
	{
		return moduleId;
	}

	protected static String getSubmittedFileName( Part part )
	{
		for( String cd : part.getHeader( "content-disposition" ).split( ";" ) )
		{
			if( cd.trim().startsWith( "filename" ) )
			{
				String fileName = cd.substring( cd.indexOf( '=' ) + 1 ).trim().replace( "\"", "" );
				fileName = fileName.substring( fileName.lastIndexOf( '/' ) + 1 ).substring( fileName.lastIndexOf( '\\' ) + 1 ); // MSIE fix.

				try
				{
					String fileName2 = new String( fileName.getBytes( "ISO-8859-1" ), "UTF-8" );

					boolean ok = false;

					char charBuffer[] = fileName2.toCharArray();
					for( char symbol : charBuffer )
					{
						int i = symbol - 'Ё';
						if( i >= 0 && i < 23 )
						{
							ok = true;
						}
						else if( symbol == '?' )
						{
							ok = false;
							break;
						}
					}

					if( ok )
					{
						fileName = fileName2;
					}
				}
				catch( UnsupportedEncodingException e )
				{
					logger.error( e.getMessage(), e );
				}

				return fileName;
			}
		}

		return null;
	}

	protected String msg( String message )
	{
		return localeBean.msg( message );
	}

	/**
	 * Запись файла в выходной поток
	 * @param dataHandler
	 * @param fileName - название файла
	 * @throws BGException
	 */
	protected void writeToOutput( DataHandler dataHandler, String fileName, boolean zipped )
		throws BGException
	{
		try
		{
			writeToOutput( dataHandler.getInputStream(), fileName, dataHandler.getContentType(), zipped );
		}
		catch( IOException e )
		{
			throw new BGException( e );
		}
	}

	/**
	 * Запись файла в выходной поток
	 * @param is
	 * @param fileName - название файла
	 * @param contentType
	 * @throws BGException
	 */
	protected void writeToOutput( InputStream is, String fileName, String contentType, boolean zipped )
		throws BGException
	{
		//задаем имя файла и тип скачиваемого файла
		//response.setContentType( "application/pdf" );
		//response.setHeader( "Content-Disposition", );

		FacesContext fc = FacesContext.getCurrentInstance();
		ExternalContext ec = fc.getExternalContext();

		ec.responseReset(); // Some JSF component library or some Filter might have set some headers in the buffer beforehand. We want to get rid of them, else it may collide.

		try
		{
			String fileNameWithExt = fileName;

			if( zipped )
			{
				fileNameWithExt = fileName + ".zip";
				contentType = "application/zip";
			}

			ec.setResponseContentType( contentType ); // Check http://www.iana.org/assignments/media-types for all types. Use if necessary ExternalContext#getMimeType() for auto-detection based on filename.
			// ec.setResponseContentLength( contentLength ); // Set it with the file size. This header is optional. It will work if it's omitted, but the download progress will be unknown.
			//ec.setResponseHeader( "Content-Disposition", "attachment; filename=\"" + fileName + "\"" ); // The Save As popup magic is done here. You can give it any file name you want, this only won't work in MSIE, it will use current request URL as file name instead.
			ec.setResponseHeader( "Content-Disposition", FileDownload.getContentDisposition( (HttpServletRequest)ec.getRequest(), fileNameWithExt, false ) );

			OutputStream output = ec.getResponseOutputStream();

			if( zipped )
			{
				ZipOutputStream zip = new ZipOutputStream( output );
				zip.putNextEntry( new ZipEntry( fileName ) );

				IOUtils.transfer( is, zip, 32 * 1024 );

				zip.flush();
				zip.close();
			}
			else
			{
				IOUtils.transfer( is, output, 32 * 1024 );
			}
		}
		catch( IOException e )
		{
			throw new BGException( e );
		}
		finally
		{
			fc.responseComplete();
		}
	}

	public <T> ThrowingSupplier<T> supEx( final ThrowingSupplier<T> supplier )
	{
		return () -> {
			try
			{
				return supplier.get();
			}
			catch( Throwable t )
			{
				if( t instanceof BGRuntimeException )
				{
					t = t.getCause();
				}

				if( t instanceof BGException || t instanceof SOAPFaultException )
				{
					String tag;

					if( t instanceof BGException )
					{
						tag = ((BGException)t).getTag();
					}
					else
					{
						tag = null;
					}

					if( t.getCause() instanceof SOAPFaultException )
					{
						t = t.getCause();
					}

					FacesContext context = FacesContext.getCurrentInstance();

					String message = null;

					if ( tag != null && tag.length() > 0 )
					{
						try
						{
							message = context.getApplication().getResourceBundle( context, "msg" ).getString( tag );
						}
						catch( Exception e )
						{
						    message = tag + " => " + e.getMessage();
						}
					}

					logger.info( String.format( "Error message: %s %s", message, t.getMessage() ) );

					context.addMessage( null, new FacesMessage( FacesMessage.SEVERITY_ERROR, "Error Message", message != null ? message : t.getMessage() ) );

					return null;
				}

				if( t instanceof RuntimeException )
				{
					throw (RuntimeException)t;
				}
				else
				{
					throw new BGRuntimeException( t );
				}
			}
		};
	}
	
	public boolean isCustomer()
	{
	    return CommonUtils.isCustomer();
	}
}
