/*
 * Copyright (c) 2016 BiTel Inc. All rights reserved.
*/
package ru.bitel.mybgbilling.kernel.common;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

import groovy.lang.Closure;
import groovy.util.ConfigObject;
import ru.bitel.bgbilling.common.BGException;
import ru.bitel.common.ParameterMap;
import ru.bitel.common.Preferences;
import ru.bitel.common.function.ThrowingFunction;
import ru.bitel.common.function.ThrowingSupplier;

public class Config
{
	private final Map<Object, Object> configMap;

	private final ThrowingSupplier<Object> argumentSupplier;
	private final ThrowingFunction<String, Object> valueProducer;

	private final ConcurrentMap<String, Object> computedMap = new ConcurrentHashMap<>();

	public Config( Map<Object, Object> configMap )
	{
		this( configMap, null, null );
	}

	public Config( Map<Object, Object> configMap, ThrowingFunction<String, Object> valueProducer, ThrowingSupplier<Object> argumentSupplier )
	{
		this.configMap = configMap;

		this.argumentSupplier = argumentSupplier;
		this.valueProducer = valueProducer;
	}

	public Object get( final String key, final Object def )
		throws BGException
	{
		return get( key, def, valueProducer, argumentSupplier );
	}

	public Object get( final String key, final Object def, final ThrowingFunction<String, Object> valueProducer, final ThrowingSupplier<Object> argumentSupplier )
		throws BGException
	{
		Object result = configMap.get( key );

		if( result instanceof Closure<?> )
		{
			final Closure<?> c = (Closure<?>)result;

			final Object value = valueProducer != null ? valueProducer.apply( key ) : null;

			if( value != null )
			{
				synchronized( c )
				{
					final Object d = c.getDelegate();
					int rs = c.getResolveStrategy();

					c.setDelegate( value );
					c.setResolveStrategy( Closure.DELEGATE_FIRST );

					if( argumentSupplier != null )
					{
						c.call( argumentSupplier.get() );
					}
					else
					{
						c.call();
					}

					result = value;

					c.setDelegate( d );
					c.setResolveStrategy( rs );
				}
			}
			else
			{
				if( argumentSupplier != null )
				{
					final Object args = argumentSupplier.get();
					if( args instanceof Object[] )
					{
						result = c.call( (Object[])argumentSupplier.get() );
					}
					else
					{
						result = c.call( argumentSupplier.get() );
					}
				}
				else
				{
					//XXX: нужен флаг, можно ли кэшировать, а может кэширование нужно не здесь
					//result = computedMap.computeIfAbsent( key, ( k ) -> c.call() );
					result = c.call();
				}
			}
		}

		if( result == null )
		{
			return def;
		}

		return result;
	}

	public Object get( final String key )
		throws BGException
	{
		return get( key, null );
	}

	public List<Config> subList( String key )
		throws BGException
	{
		return subList( key, valueProducer, argumentSupplier );
	}

	@SuppressWarnings("unchecked")
	public List<Config> subList( final String key, final ThrowingFunction<String, Object> valueProducer, final ThrowingSupplier<Object> argumentSupplier )
		throws BGException
	{
		if( argumentSupplier == null )
		{
			return (List<Config>)computedMap.computeIfAbsent( key, (ThrowingFunction<String, Object>)( k ) -> subList0( k, valueProducer, argumentSupplier ) );
		}
		else
		{
			return subList0( key, valueProducer, argumentSupplier );
		}
	}

	private List<Config> subList0( final String key, final ThrowingFunction<String, Object> valueProducer, final ThrowingSupplier<Object> argumentSupplier )
		throws BGException
	{
		final List<ConfigObject> result = new ArrayList<>( 5 );

		final Object value = configMap.get( key );

		if( value instanceof List )
		{
			@SuppressWarnings("unchecked")
			List<Object> list = (List<Object>)value;

			for( Object o : list )
			{
				if( o instanceof Closure<?> )
				{
					final Closure<?> c = (Closure<?>)o;

					final ConfigObject co = new ConfigObject();
					c.setDelegate( co );
					c.setResolveStrategy( Closure.DELEGATE_FIRST );

					c.call();

					result.add( co );
				}
			}
		}
		else if( value instanceof Closure<?> )
		{
			final Closure<?> c = (Closure<?>)result;

			final ConfigObject co = new ConfigObject();
			c.setDelegate( co );
			c.setResolveStrategy( Closure.DELEGATE_FIRST );

			c.call();

			result.add( co );
		}
		else if( value instanceof ConfigObject )
		{
			result.add( (ConfigObject)value );
		}

		final List<Config> rr = new ArrayList<>( result.size() );
		for( ConfigObject co : result )
		{
			rr.add( new Config( co.flatten(), valueProducer, argumentSupplier ) );
		}

		return rr;
	}

	public ParameterMap getParameterMap()
	{
		Map<String, String> map = new HashMap<>();

		for( Map.Entry<Object, Object> e : configMap.entrySet() )
		{
			map.put( String.valueOf( e.getKey() ), String.valueOf( e.getValue() ) );
		}

		return new Preferences( map );
	}
}
