/*
 * Copyright (c) 2016 BiTel Inc. All rights reserved.
*/
package ru.bitel.mybgbilling.kernel.common;

import java.io.ByteArrayInputStream;
import java.io.Serializable;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Currency;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.TimeUnit;
import java.util.function.Function;
import java.util.stream.Collectors;

import javax.annotation.PostConstruct;
import javax.el.LambdaExpression;
import javax.enterprise.context.ApplicationScoped;
import javax.enterprise.inject.Instance;
import javax.faces.context.ExternalContext;
import javax.faces.context.FacesContext;
import javax.inject.Inject;
import javax.inject.Named;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import groovy.lang.Closure;
import groovy.lang.GroovyShell;
import groovy.lang.MissingMethodException;
import groovy.util.ConfigObject;
import ru.bitel.bgbilling.common.BGException;
import ru.bitel.bgbilling.common.BGRuntimeException;
import ru.bitel.bgbilling.kernel.admin.plugincfg.common.PluginItem;
import ru.bitel.bgbilling.kernel.admin.plugincfg.common.PlugincfgService;
import ru.bitel.bgbilling.kernel.contract.api.common.bean.Contract;
import ru.bitel.bgbilling.kernel.contract.api.common.bean.parameter.ContractParameterGroupAttr;
import ru.bitel.bgbilling.kernel.module.common.bean.BGModule;
import ru.bitel.bgbilling.kernel.module.common.service.ModuleService;
import ru.bitel.common.ParameterMap;
import ru.bitel.common.Preferences;
import ru.bitel.common.Utils;
import ru.bitel.common.function.Expirable;
import ru.bitel.common.function.ThrowingFunction;
import ru.bitel.common.function.ThrowingSupplier;
import ru.bitel.common.groovy.GroovyBeanBuilder;
import ru.bitel.mybgbilling.kernel.menu.MenuItem;
import ru.bitel.mybgbilling.kernel.ws.CustomerBean;
import ru.bitel.oss.kernel.entity.common.bean.EntitySpecAttrType;

@Named
@ApplicationScoped
public class Configuration
{
	private static final Logger logger = LogManager.getLogger();

	/**
	 * Режим аутентификации.
	 */
	public static class AuthenticationMode
		implements Serializable
	{
        private static final long serialVersionUID = 627384943344506058L;
        
        private String title;
        
        /**
         * Режим аутентификации, см. использование {@link #getMode()}.
         */
        private String mode;
		private String module;
		private int moduleId;
		
		private int parameterId;
		private ContractParameterGroupAttr parameterSpec;
		
		/**
		 * Роль, если указано "unauthCustomer", то после аутентификации у абонента будет ограниченный доступ.
		 */
		private String role;

		/**
		 * Фильтр, указывающий, подходит ли данный режим для договора. Применяется после поиска договора - даже если договор найден в данном режиме,
		 * все равно произойдет ошибка "Договор не найден".
		 */
		private Closure<Boolean> filter;
		
		private boolean usernameIgnoreCase = true;
		
        @SuppressWarnings( "unchecked" )
        private Closure<String> username = (Closure<String>)new GroovyShell().evaluate( "{ s -> s.replaceAll( /^8(.+)$/,'7$1' ) }" );

		/**
		 * Режим аутентификации, см. использование данного метода.
		 * @return
		 */
		public String getMode()
		{
			return mode;
		}

		public String getModule()
		{
			return module;
		}

		public int getModuleId()
		{
			return moduleId;
		}

		public Closure<Boolean> getFilter()
		{
			return filter;
		}

		public void setFilter( Closure<Boolean> filter )
		{
			this.filter = filter;
		}

		public String getRole()
		{
			return role;
		}

		public boolean isUsernameIgnoreCase()
		{
			return usernameIgnoreCase;
		}

        public int getParameterId()
        {
            return parameterId;
        }

        public int getParameterType()
        {
            return parameterSpec != null ? parameterSpec.getType() : EntitySpecAttrType.TEXT.getCode();
        }

        public String getParameterTypeName()
        {
            return parameterSpec != null ? parameterSpec.getTypeName() : ContractParameterGroupAttr.getTypeName( EntitySpecAttrType.TEXT.getCode() );
        }
        
        public ContractParameterGroupAttr getParameterSpec()
        {
            return parameterSpec;
        }

        public Closure<String> getUsername()
        {
            return username;
        }

        public String getTitle()
        {
            return title;
        }
	}

	public static class Authentication
	{
		// Кол-во ошибок аутентификации, после которого будет отображаться captcha для этого логина
		private int captchaLoginErrorCount = 5;

		// Кол-во ошибок аутентификации, после которого будет отображаться captcha для хоста
		private int captchaHostErrorCount = 20;

		// Кол-во ошибок аутентификации, после которых будут заблокированы попытки этого хоста
		private int blockHostErrorCount = 30;

		private List<AuthenticationMode> modes = new ArrayList<>();

		public int getCaptchaLoginErrorCount()
		{
			return captchaLoginErrorCount;
		}

		public int getCaptchaHostErrorCount()
		{
			return captchaHostErrorCount;
		}

		public int getBlockHostErrorCount()
		{
			return blockHostErrorCount;
		}

		public List<AuthenticationMode> getModes()
		{
			return modes;
		}
	}
	
	/**
	 * Базовый URL ЛК.
	 */
	private String baseUrl = null;
	private URI baseURI = null;

	/**
	 * Используемая тема, см. WebContext/resources/theme
	 */
	private String theme = "yeti";

	/**
	 * Список модулей биллинга.
	 */
	private Expirable<List<BGModule>> moduleList;
	
	/**
	 * Список плагинов биллинга.
	 */
	private Expirable<Map<String, PluginItem>> pluginMap;

	private String currencySymbol = "";
	private String currencyCode = "RUB";

	private String hostHttpRequestHeader = null;
	private String portHttpRequestHeader = null;
	
	/**
	 * Начальная страница, например, "kernel/news".
	 */
	private String startPage;

	/**
	 * Полный groovy-объект конфигурации (обработанный mybgbilling-conf.groovy).
	 */
	private ConfigObject configObject;

	@Inject
	private Instance<CustomerBean> customerBean;

	@Inject
	private Instance<ConfigurationCacheBean> configurationCacheBean;

	ConfigObject getConfigObject()
	{
		return configObject;
	}

	/**
	 * "Плоский" конфиг с ключами вида "content.kernel.customerTitle".
	 */
	private Map<Object, Object> configMap;

	/**
	 * Обертка вокруг "плоского" конфига.
	 */
	private Config config;

	private ThrowingSupplier<Object> configArgumentSupplier = () -> {

		final CustomerBean customerBean = this.customerBean.get();
		return customerBean.getContract();
	};

	private Currency currency;
	
    private Map<Integer, List<Integer>> activateModuleActions = new HashMap<>();

    public Map<Integer, List<Integer>> getActivateModuleActions()
    {
        return activateModuleActions;
    }

    @PostConstruct
	private void init()
	{
		try
		{
		    // статические методы данного класса будут доступны для вызова в файле конфигурации
			GroovyBeanBuilder.addCommonStaticStars( "ru.bitel.mybgbilling.kernel.common.utils.CommonUtils" );
			
			GroovyBeanBuilder bb = new GroovyBeanBuilder();
			bb.setClassLoader( this.getClass().getClassLoader() );
			
			// проставляем объектом context файла конфигурации данный объект Configuration
			bb.setInitialVariable( "context", this );

			// мапим классы - объекты будут созданы при парсинге файла
			bb.setMappedClass( "authentication", Authentication.class );
			bb.setMappedClass( "authenticationMode", AuthenticationMode.class );

			ExternalContext externalContext = FacesContext.getCurrentInstance().getExternalContext();
			
			// обработка файла конфигурации. чуть ниже идет дополнительная постобработка
			this.configObject = bb.load( externalContext.getResourceAsStream( "/WEB-INF/mybgbilling-conf.groovy" ), "mybgbilling-conf.groovy" );

			// получаем "плоский" конфиг с ключами вида "content.kernel.customerTitle"
			this.configMap = this.configObject.flatten();
			// удаляем параметры подключения к биллингу, они здесь лишние (для получения этих параметров можно использовать configObject)
			this.configMap.entrySet().removeIf( ( e ) -> e.getKey() instanceof String && ((String)e.getKey()).startsWith( "bgbilling." ) );
		}
		catch( Exception e )
		{
			logger.error( "Error while parsing mybgbilling-conf.groovy" );
			logger.error( e.getMessage(), e );
		}

		try
		{
			this.config = new Config( configMap );
		}
		catch( Exception e )
		{
			logger.error( e.getMessage(), e );
		}

		if( Utils.isBlankString( hostHttpRequestHeader ) )
		{
			hostHttpRequestHeader = null;
		}
		
        if( Utils.isBlankString( portHttpRequestHeader ) )
        {
            portHttpRequestHeader = null;
        }

		moduleList = Expirable.of( () -> {

			MyApplicationBean application = (MyApplicationBean)FacesContext.getCurrentInstance()
																		   .getELContext().getELResolver()
																		   .getValue( FacesContext.getCurrentInstance().getELContext(), null, "myApplicationBean" );

			ModuleService moduleService = application.getWsClient().getPort( ModuleService.class, 0 );
			return moduleService.moduleList();

		}, 6, TimeUnit.MINUTES );
		
		pluginMap = Expirable.of( () -> {

			MyApplicationBean application = (MyApplicationBean)FacesContext.getCurrentInstance()
																		   .getELContext().getELResolver()
																		   .getValue( FacesContext.getCurrentInstance().getELContext(), null, "myApplicationBean" );

			PlugincfgService plugincfgService = application.getWsClient().getPort( PlugincfgService.class, 0 );
			return Utils.maskNull( plugincfgService.getPlugins() ).stream().collect( Collectors.toMap( a -> a.getName(), a -> a ) );

		}, 6, TimeUnit.MINUTES );

		if( Utils.isBlankString( currencySymbol ) && Utils.isBlankString( currencyCode ) )
		{
			currencyCode = "RUB";
			currencySymbol = "";
		}

		currency = Currency.getInstance( currencyCode );

		ExternalContext externalContext = FacesContext.getCurrentInstance().getExternalContext();
		externalContext.getApplicationMap().put( "currencySymbol", currencySymbol );
		externalContext.getApplicationMap().put( "curSym", currencySymbol );
		externalContext.getApplicationMap().put( "currencyCode", currencyCode );
		externalContext.getApplicationMap().put( "curCode", currencyCode );

		try
		{
			this.authentication = (Authentication)this.get( "authentication" );

			if( this.authentication == null )
			{
				this.authentication = new Authentication();
			}

			// если режимы аутентификации не указаны, то создаем один режим - по договору
			if( this.authentication.modes == null || this.authentication.modes.size() == 0 )
			{
				AuthenticationMode mode = new AuthenticationMode();
				mode.mode = "contract";

				this.authentication.modes = new ArrayList<>( 1 );
				this.authentication.modes.add( mode );
			}
			else
			{
                for( AuthenticationMode mode : this.authentication.modes )
                {
                    if( mode.role == null )
                    {
                        mode.role = "customer";
                    }
                }
			}
		}
		catch( Exception e )
		{
			logger.error( e.getMessage(), e );
		}
	}

	private Authentication authentication;

	public Authentication getAuthentication()
	{
		return authentication;
	}

	public String getCurrencyCode()
	{
		return currencyCode;
	}

	public Currency getCurrency()
	{
		return currency;
	}

    public URI getBaseURI()
    {
        if( baseUrl == null )
        {
            return null;
        }

        if( baseURI == null )
        {
            if( baseUrl.endsWith( "/" ) )
            {
                baseUrl = baseUrl.substring( 0, baseUrl.length() - 1 );
            }

            try
            {
                baseURI = new URI( baseUrl );
            }
            catch( URISyntaxException e )
            {
                baseUrl = null;
                logger.error( e.getMessage(), e );
            }
        }

        return baseURI;
    }

    public String getTheme()
	{
		return theme;
	}

	public String getStartPage()
	{
		return startPage;
	}

	public List<BGModule> getModuleList()
	{
		return moduleList != null ? moduleList.get() : null;
	}

	public String getHostHttpRequestHeader()
	{
		return hostHttpRequestHeader;
    }

    public String getPortHttpRequestHeader()
    {
        return portHttpRequestHeader;
    }

    /**
     * Получение значения из конфигурации по ключу, например:
     * <pre>configuration.get('content.kernel.status.statusChange', true)</pre>
     * Если значение является лямбдой (замыканием), то происходит её вызов c передачей агрумента contract, а результат кэшируется для сессии.
     * @param key ключ
     * @param def значение по умолчанию
     * @return
     * @throws BGException
     */
    public Object get( final String key, final Object def )
		throws BGException
	{
		return configurationCacheBean.get().computeIfAbsent( key, (ThrowingFunction<String, Object>)( k ) -> config.get( k, def, null, configArgumentSupplier ) );
		//return config.get( key, def, null, configArgumentSupplier );
	}

	public Object get( final String key, final ThrowingSupplier<Object> argumentSupplier, final Object def )
		throws BGException
	{
		try
		{
			return configurationCacheBean.get().computeIfAbsent( key, (ThrowingFunction<String, Object>)( k ) -> config.get( k, def, null, argumentSupplier ) );
		}
		catch( MissingMethodException ex )
		{
			logger.error( ex.getMessage(), ex );
		}

		return def;
	}

	public Object invoke( final String key, final ThrowingSupplier<Object> argumentSupplier, final Object def )
		throws BGException
	{
		try
		{
			return config.get( key, def, null, argumentSupplier );
		}
		catch( MissingMethodException ex )
		{
			logger.error( ex.getMessage(), ex );
		}

		return def;
	}

    /**
     * Получение значения из конфигурации по ключу, например:
     * <pre>
     * configuration.invoke('content.kernel.cards.filter', ()->[customerBean.contract, a], true )
     * configuration.invoke('aa.bb.cc', arg, false )</pre>
     * Если значение является лямбдой (замыканием), то происходит её вызов с агрументом, при этом в отличие от get, результат <b>не</b> кэшируется для сессии.<br/>
     * В EL не поддерживается переменное кол-во агрументов (Object...), поэтому при передаче нескольких аргументах для лямбды,
     * необходимо их передать в лямбде, например: ()->[customerBean.contract, a] 
     * @param key ключ
     * @param argument один агрумент или лямбда, возвращающая массив агрументов: ()->[customerBean.contract, a]
     * @param def значение по умолчанию
     * @return
     * @throws BGException
	 */
	public Object invoke( final String key, final Object argument, final Object def )
		throws BGException
	{
		Object result;

		if( argument instanceof LambdaExpression )
		{
			result = invoke( key, () -> {

				Object arg = ((LambdaExpression)argument).invoke();
				if( arg instanceof ArrayList<?> )
				{
					return ((List<?>)arg).toArray();
				}
				else
				{
					return arg;
				}

			}, DEF );
		}
		else
		{
			result = invoke( key, () -> argument, DEF );
		}

		if( result == DEF )
		{
			if( def instanceof LambdaExpression )
			{
				return ((LambdaExpression)def).invoke();
			}
			else
			{
				return def;
			}
		}

		return result;
	}

	private static final Object DEF = new Object();

    /**
     * Получение значения из конфигурации по ключу, например:
     * <pre>
     * configuration.invoke('content.kernel.cards.filter', ()->[customerBean.contract, a], true )
     * configuration.invoke('aa.bb.cc', arg, false )</pre>
     * Если значение является лямбдой (замыканием), то происходит её вызов c передачей аргумента, результат кэшируется для сессии.<br/>
     * В EL не поддерживается переменное кол-во агрументов (Object...), поэтому при передаче нескольких аргументах для лямбды,
     * необходимо их передать в лямбде, например: ()->[customerBean.contract, a] 
     * @param key ключ
     * @param argument один агрумент или лямбда, возвращающая массив агрументов: ()->[customerBean.contract, a]
     * @param def значение по умолчанию
     * @return
     * @throws BGException
     */
	public Object get( final String key, final Object argument, final Object def )
		throws BGException
	{
		Object result;

		if( argument instanceof LambdaExpression )
		{
			result = get( key, () -> {

				Object arg = ((LambdaExpression)argument).invoke();
				if( arg instanceof ArrayList<?> )
				{
					return ((List<?>)arg).toArray();
				}
				else
				{
					return arg;
				}

			}, DEF );
		}
		else
		{
			result = get( key, () -> argument, DEF );
		}

		if( result == DEF )
		{
			if( def instanceof LambdaExpression )
			{
				return ((LambdaExpression)def).invoke();
			}
			else
			{
				return def;
			}
		}

		return result;
	}

	/**
	 * Получение переменной из конфигурации. Если в конфигурации нет данной переменной, то вернется null.
	 * @param key ключ
	 * @return
	 * @throws BGException
	 */
    public Object get( final String key )
        throws BGException
    {
        // этот метод может быть вызван еще при парсинге конфига, когда в GroovyBeanBuilder вызывается InvokerHelper.getProperty()
        if( config == null )
        {
            return null;
        }

        return configurationCacheBean.get().computeIfAbsent( key, (ThrowingFunction<String, Object>)( k ) -> config.get( k ) );
    }

	public List<Config> subList( String key )
		throws BGException
	{
		return this.subList( key, null );
	}

	public List<Config> subList( String key, ThrowingSupplier<Object> argumentSupplier )
		throws BGException
	{
		return config.subList( key, null, argumentSupplier );
	}

	public ParameterMap getParameterMap()
	{
		Map<String, String> map = new HashMap<>();

		for( Map.Entry<Object, Object> e : configMap.entrySet() )
		{
			map.put( String.valueOf( e.getKey() ), String.valueOf( e.getValue() ) );
		}

		return new Preferences( map );
	}

	public static class MenuFactory
	{
		public Closure<?> createMenu;
	}

	volatile Function<Contract, MenuItem> menuFactory;

	public Function<Contract, MenuItem> getMenuFactory()
	{
		if( this.menuFactory != null )
		{
			return this.menuFactory;
		}

		try
		{
			ExternalContext externalContext = FacesContext.getCurrentInstance().getExternalContext();
			String script = new String( Utils.readByBlock( externalContext.getResourceAsStream( "/WEB-INF/mybgbilling-menu.groovy" ) ) );

			//limitChange = { contract -> (contract.balanceMode == (byte)1) }
			script = "menuFactory { createMenu = { contract -> " + script + "\n}}";

			GroovyBeanBuilder bb = new GroovyBeanBuilder();
			bb.setClassLoader( this.getClass().getClassLoader() );
			//bb.setInitialVariable( "menu", menu );

			MenuFactory menuFactory = new MenuFactory();
			bb.setInitialVariable( "menuFactory", menuFactory );

//			bb.getBinding().setProperty( "contract", customerBean.getContract() );

			bb.setMappedClass( "menu", MenuItem.class );

			//bb.load( externalContext.getResourceAsStream( "/WEB-INF/mybgbilling-menu.groovy" ) );
			try
			{
				bb.load( new ByteArrayInputStream( script.getBytes( "UTF-8" ) ), "mybgbilling-menu.groovy" );
			}
			catch( BGException ex )
			{
				logger.error( "Error while parsing mybgbilling-menu.groovy" );
				throw ex;
			}
			catch( RuntimeException ex )
			{
				logger.error( "Error while parsing mybgbilling-menu.groovy" );
				throw ex;
			}

			bb = null;

			this.menuFactory = ( contract ) -> {
				return (MenuItem)menuFactory.createMenu.call( new Object[] { contract } );
			};
		}
		catch( Exception ex )
		{
			throw new BGRuntimeException( ex );
		}

		return this.menuFactory;
	}
	
	/**
	 * Проверка включен ли модуль в сервере.
	 * @param plugin имя плагина, без префикса "ru.bitel.bgbilling.plugins."
	 * @return
	 */
	public boolean isPluginEnabled( String plugin )
	{
		try
		{
			PluginItem pluginItem = pluginMap.get().get( "ru.bitel.bgbilling.plugins." + plugin );
			return pluginItem != null ? pluginItem.isEnabled() : false;
		}
		catch( Exception ex )
		{
			logger.error( ex.getMessage(), ex );
		}

		return false;
	}
	
	public Expirable<Map<String, PluginItem>> getPluginMap()
	{
		return pluginMap;
	}
	
	public static void main( String[] args )  
	{
	    Closure<String> username = (Closure<String>)new GroovyShell().evaluate( "{ s -> s.replaceAll( /^8(.+)$/ ){ all, m -> '7'+m } }" );
	    username = (Closure<String>)new GroovyShell().evaluate( "{ s -> s.replaceAll( '^8(.+)$', '7$1\' ) }" );
	    username = (Closure<String>)new GroovyShell().evaluate( "{ s -> s.replaceAll( /^8(.+)$/, '7$1\' ) }" );
	    
	    System.out.println( username.call( "8917401" ) );
	}
}
