/*
 * Copyright (c) 2016 BiTel Inc. All rights reserved.
*/
package ru.bitel.mybgbilling.kernel.common;

import java.io.ByteArrayOutputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.SecureRandom;
import java.time.LocalDate;
import java.time.format.DateTimeFormatter;
import java.util.concurrent.atomic.AtomicLong;

import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import javax.enterprise.context.ApplicationScoped;
import javax.enterprise.inject.Produces;
import javax.inject.Inject;
import javax.inject.Named;
import javax.servlet.http.HttpServletRequest;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.w3c.dom.Document;

import groovy.util.ConfigObject;
import ru.bitel.bgbilling.common.BGException;
import ru.bitel.bgbilling.common.BGRuntimeException;
import ru.bitel.common.Utils;
import ru.bitel.common.XMLUtils;
import ru.bitel.common.groovy.GroovyBeanBuilder;
import ru.bitel.common.util.TimeoutMap;
import ru.bitel.mybgbilling.kernel.common.Configuration.Authentication;
import ru.bitel.mybgbilling.kernel.common.TransferData.HTTPDataSource;
import ru.bitel.mybgbilling.kernel.ws.WSClient;
import ru.bitel.bgbilling.common.bean.Request;

@Named
@ApplicationScoped
public class MyApplicationBean
{
	private static final Logger logger = LogManager.getLogger();

	@Inject
	private Configuration configuration;

	private TimeoutMap<String, AtomicLong> ipCheckMap;
	private TimeoutMap<String, AtomicLong> loginCheckMap;

	private TimeoutMap<String, Integer> tokenAuthMap;

	private final SecureRandom random = new SecureRandom();

	private final CaptchaRenderer captchaRenderer = new CaptchaRenderer();

	private WSClient wsClient;

	private URL _url;

	private String _user;
	private String _password;

	@PostConstruct
	private void init()
	{
		logger.info( "MyApplicationBean.init" );

		GroovyBeanBuilder.addCommonStaticStars( "ru.bitel.mybgbilling.kernel.common.utils.CommonUtils" );

		ipCheckMap = new TimeoutMap<>( "ipCheckMap" );
		loginCheckMap = new TimeoutMap<>( "loginCheckMap" );

		tokenAuthMap = new TimeoutMap<>( "tokenAuthMap" );

		String url = "http://127.0.0.1:8080/bgbilling/executer";

		ConfigObject configObject = configuration.getConfigObject();

		try
		{
			ConfigObject conf = (ConfigObject)configObject.getProperty( "bgbilling" );

			url = (String)conf.getOrDefault( "url", url );
		}
		catch( BGRuntimeException e )
		{
			logger.error( e.getMessage(), e );
		}

		String user = "admin";
		String password = "admin";

		try
		{
			ConfigObject conf = (ConfigObject)configObject.getProperty( "bgbilling" );

			user = (String)conf.getOrDefault( "user", user );
			password = (String)conf.getOrDefault( "password", password );
		}
		catch( BGRuntimeException e )
		{
			logger.error( e.getMessage(), e );
		}
		
        url = System.getProperty( "my.bgbilling.url", url );
        user = System.getProperty( "my.bgbilling.user", user );
        password = System.getProperty( "my.bgbilling.password", password );

		wsClient = new WSClient( url, user, password, true, true );

		try
		{
			_url = new URL( url );
		}
		catch( MalformedURLException e )
		{
			logger.error( e.getMessage(), e );
		}

		_user = user;
		_password = password;
	}

	@PreDestroy
	private void destroy()
	{
		logger.info( "MyApplicationBean.destroy" );

		this.ipCheckMap.shutdown();
		this.loginCheckMap.shutdown();
		this.tokenAuthMap.shutdown();
	}

	public String generateToken( int contractId )
	{
		String token = generateToken();
		tokenAuthMap.put( token, contractId, System.currentTimeMillis() + 20000L );

		return token;
	}

	public int getContractIdByToken( String token )
	{
		Integer result = tokenAuthMap.remove( token );
		return result != null ? result : 0;
	}

	private String generateToken()
	{
		int length = 64;

		String chars = "1234567890ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
		int charsLength = chars.length();

		StringBuilder sb = new StringBuilder();

		for( int i = 0; i < length; i++ )
		{
			sb.appendCodePoint( chars.codePointAt( random.nextInt( charsLength ) ) );
		}

		return sb.toString();
	}

	public void processLoginFailure( final String remoteAddr, final String username )
	{
		if( Utils.isBlankString( remoteAddr ) || "127.0.0.1".equals( remoteAddr ) )
		{
			logger.warn( "Remote addr is: " + remoteAddr );
		}
		else
		{
			AtomicLong ipCount = ipCheckMap.get( remoteAddr );
			if( ipCount == null )
			{
				AtomicLong newIpCount = new AtomicLong();
				ipCount = ipCheckMap.putIfAbsent( remoteAddr, newIpCount, System.currentTimeMillis() + 3600000 );
				if( ipCount == null )
				{
					ipCount = newIpCount;
				}
			}

			ipCount.incrementAndGet();
		}

		if( Utils.notBlankString( username ) )
		{
			AtomicLong loginCount = loginCheckMap.get( username );
			if( loginCount == null )
			{
				AtomicLong newLoginCount = new AtomicLong();
				loginCount = loginCheckMap.putIfAbsent( username, newLoginCount, System.currentTimeMillis() + 600000 );
				if( loginCount == null )
				{
					loginCount = newLoginCount;
				}
			}

			loginCount.incrementAndGet();
		}
	}

	public void processLoginOk( final String remoteAddr, final String username )
	{
		loginCheckMap.remove( username );

		if( Utils.isBlankString( remoteAddr ) || "127.0.0.1".equals( remoteAddr ) )
		{
			logger.warn( "Remote addr is: " + remoteAddr );
		}
		else
		{
			final AtomicLong ipCount = ipCheckMap.get( remoteAddr );
			if( ipCount != null )
			{
				ipCount.decrementAndGet();
			}
		}
	}

	public boolean needCaptcha( final String remoteAddr, final String username )
	{
		final Authentication authentication = configuration.getAuthentication();

		if( Utils.notBlankString( username ) )
		{
			final AtomicLong loginCount = loginCheckMap.get( username );
			if( loginCount != null )
			{
				if( loginCount.get() >= authentication.getCaptchaLoginErrorCount() )
				{
					return true;
				}
			}
		}

		if( Utils.isBlankString( remoteAddr ) || "127.0.0.1".equals( remoteAddr ) )
		{
			logger.warn( "Remote addr is: " + remoteAddr );
		}
		else
		{
			final AtomicLong ipCount = ipCheckMap.get( remoteAddr );
			if( ipCount != null )
			{
				if( ipCount.get() >= authentication.getCaptchaHostErrorCount() )
				{
					return true;
				}
			}
		}

		return false;
	}

	public CaptchaRenderer getCaptchaRenderer()
	{
		return captchaRenderer;
	}

	public String getRemoteAddr( final HttpServletRequest req )
	{
		final String header = configuration.getHostHttpRequestHeader();
		if( header == null )
		{
			return req.getRemoteAddr();
		}

		String result = req.getHeader( header );
		if( result != null )
		{
			return result;
		}

		return req.getRemoteAddr();
	}
	
    public int getRemotePort( final HttpServletRequest req )
    {
        final String header = configuration.getPortHttpRequestHeader();
        if( header == null )
        {
            return req.getRemotePort();
        }

        String result = req.getHeader( header );
        if( result != null )
        {
            int port = Utils.parseInt( result, -1 );
            if( port >= 0 )
            {
                return port;
            }
        }

        return req.getRemotePort();
    }

	WSClient getWsClient()
	{
		return wsClient;
	}

	@Produces
	HttpRequest newHttpRequest()
	{
		return new HttpRequest();
	}

	public class HttpRequest
	{
		HttpRequest()
		{}

		private Request newRequest( String module, int moduleId, String action, int contractId, Object... args )
		{
			Request request = new Request();
			request.setModule( module );
			request.setModuleId( moduleId );
			request.setAction( action );
			request.setContractId( contractId );

			for( int i = 0; i < args.length; i += 2 )
			{
				String value;

				if( args[i + 1] instanceof LocalDate )
				{
					DateTimeFormatter formatter = DateTimeFormatter.ofPattern( "dd.MM.yyyy" );
					value = formatter.format( (LocalDate)args[i + 1] );
				}
				else if( args[i + 1] == null )
				{
					value = "";
				}
				else
				{
					value = String.valueOf( args[i + 1] );
				}

				request.setAttribute( (String)args[i], value );
			}
			return request;
		}

		public Document requestXml( String module, int moduleId, String action, int contractId, Object... args )
			throws BGException
		{
			try
			{
				Request request = newRequest( module, moduleId, action, contractId, args );

				TransferData td = new TransferData( _url );
				td.setUser( _user, _password );
				td.postData( request );

				Document result = td.getDocument();

				ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
				XMLUtils.serialize( result, outputStream, "UTF-8" );
				logger.info( "xmlResponse: " + new String( outputStream.toByteArray() ) );

				return result;
			}
			catch( Throwable e )
			{
				logger.error( e.getMessage(), e );

				throw new BGException( e.getMessage(), e );
			}
		}

		public HTTPDataSource request( String module, int moduleId, String action, int contractId, Object... args )
			throws BGException
		{
			try
			{
				Request request = newRequest( module, moduleId, action, contractId, args );

				TransferData td = new TransferData( _url );
				td.setUser( _user, _password );

				HTTPDataSource result = td.getDataSource( request );

				logger.info( "dataSourceResponse" );

				return result;
			}
			catch( Throwable e )
			{
				logger.error( e.getMessage(), e );

				throw new BGException( e.getMessage(), e );
			}
		}
	}
}
