package ru.bitel.mybgbilling.kernel.common.inject;

import java.io.File;
import java.io.IOException;
import java.io.StringWriter;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.List;
import java.util.Locale;
import java.util.UUID;

import javax.tools.Diagnostic;
import javax.tools.DiagnosticListener;
import javax.tools.JavaCompiler;
import javax.tools.JavaCompiler.CompilationTask;
import javax.tools.JavaFileObject;
import javax.tools.StandardJavaFileManager;
import javax.tools.ToolProvider;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import ru.bitel.bgbilling.common.BGException;
import ru.bitel.bgbilling.kernel.dynamic.common.bean.CompilationMessage;
import ru.bitel.bgbilling.kernel.dynamic.common.bean.CompilationResult;
import ru.bitel.common.model.Pair;

/** ru.bitel.bgbilling.kernel.dynamic.server.CompilerWrapper - примерный его клон. Ни наследование ни импорт его тут не подошел. */
public final class CompilerBeanWrapper
{
	private static final Logger logger = LogManager.getLogger();
	private File outputdirRoot;
	private File outputdir;
	private File srcdir;

	public CompilerBeanWrapper( File srcdir, File outputdirRoot )
	{
		this.srcdir = srcdir;
		this.outputdirRoot = outputdirRoot;
	}

	/**
	 * Компилирует поданные исходники. Сюда должны попадать только java-файлы, не ресурсы.
	 * @param srcFiles список полных путей к файлам
	 * @return пара: результат компиляции и список скомпилированных файлов
	 * @throws BGException 
	 * @throws IOException 
	 */
	public Pair<CompilationResult, List<CompiledUnit>> compile( List<String> srcFiles , ClassLoader classLoader ) throws BGException, IOException
	{
		//берем компилятор
		JavaCompiler compiler = ToolProvider.getSystemJavaCompiler();
		if( compiler == null )
		{
			throw new BGException( "System error. No Java compiler found on this platform!" );
		}

		//добавляем диагностик листенера
		DiagnosticListenerImpl listener = new DiagnosticListenerImpl();
		//берем файл-менеджера
		StandardJavaFileManager fileManager = compiler.getStandardFileManager( listener, null, null );
		
		//получаем ява-объекты из сорцов
		Iterable<? extends JavaFileObject> fileObjects = fileManager.getJavaFileObjectsFromStrings( srcFiles );
		
		//говорим куда и как конпейлять
		// устанавливаем временную папку для скомпилированных файлов с уникальным именем
		String outputdirPath = outputdirRoot.getAbsoluteFile() + File.separator + UUID.randomUUID().toString();
		outputdir = new File( outputdirPath );
		if(!outputdir.mkdirs())
		{
			// должно 1) не быть каталога до этого (уникальный путь у нас) и 2) должно обязательно создаться, так что проверка на true корректная
			// проверка на всякий случай, у клиента почему-то mkdir() молча ничего не делал (кажется), хотя права итд были, заодно правильнее mkdirs()
			throw new BGException("error create javac-output directory " + outputdirPath);
		}
		
		// мы хотим конпилять в спецификации под которой скопмилен основной сервер, и исходники в этой же видеть расчитываем. это правильно.
		// потом можно передать будет через -source и -target. Но способа корректно найти это, походу, нет.
		
		String[] options = new String[] { "-classpath", getClasspath( classLoader ),
			                         "-sourcepath", srcdir.getAbsolutePath(),
			                         "-encoding", "UTF-8",
			                         "-d", outputdir.getAbsolutePath(),
			                         "-verbose", "-Xlint:deprecation",
			                         "-g:vars,lines,source",  //расширенная debug-информация
									 "-parameters"};  // названия параметров у методов внутри class-файла.

		//создаем задание на компиляцию
		StringWriter sw = new StringWriter();
		CompilationTask task = compiler.getTask( sw, fileManager, listener, Arrays.asList( options ), null, fileObjects );
		
//		List<String> modules = new ArrayList<>();
//		String moduleName = classLoader.getParent().getClass().getModule().getName();
//		modules.add( moduleName ); 
//		task.addModules( modules );

		//конпейлируем!
		boolean callResult = task.call();
		CompilationResult result = listener.getCompilationResult();
		
		if( !callResult )
		{
		    StringBuilder text = new StringBuilder();
            for( CompilationMessage msg : result.getErrors() )
            {
                text.append( msg.getMessage() + " at " + msg.getSource() + " on ");
                text.append( msg.getLine() + ":" + msg.getColumn() + "\n" );
            }
            throw new BGException( text.toString() );
		}
		
		return new Pair<CompilationResult, List<CompiledUnit>>( result, compiledFiles( outputdir ) );
	}
	
	private static String classpath = null;
	private static String getClasspath( ClassLoader classLoader ) throws IOException
    {
		if( classpath != null ) {
			return classpath;
		}
	    
	    StringBuilder sb = new StringBuilder();
	    Enumeration<URL> resources = classLoader.getResources("/");
        while (resources.hasMoreElements() ) {
            URL url = resources.nextElement();
            sb.append( url.getFile().replaceAll("%20", " ") ).append(File.pathSeparator);
        }
	    return classpath = sb.toString();
    }

	/**
	 * Получает список объектов с информацией о скомпилированных файлах, лежащих в указанной директории. 
	 * @param directory вынесено в параметр только для рекурсивного вызова, на самом деле просто outputdir подразумевается изначально.
	 * @return
	 */
	private List<CompiledUnit> compiledFiles( File directory )
	{
		List<CompiledUnit> result = new ArrayList<CompiledUnit>();
		
		for( File tmpdirFile : directory.listFiles() )
		{
			// если файл (а не папка), добавляем его данные в список
			if( tmpdirFile.isFile() )
			{
				File compiledFile = tmpdirFile;
				String filePath = compiledFile.getAbsolutePath();
				
				String className = filePath.substring( outputdir.getAbsolutePath().length() + 1 ).replace( File.separatorChar, '.' );
				className = className.substring( 0, className.length() - 6 ); // 6 == ".class".length()
				
				if( logger.isDebugEnabled() )
				{
					logger.debug( "Found compiled unit: " + className + " with file " + compiledFile.getAbsolutePath() );
				}
				
				CompiledUnit unit = new CompiledUnit();
				unit.className = className;
				unit.classFile = compiledFile;
				unit.srcFile = getClassFile( className );
				result.add( unit );
			}
			else if( tmpdirFile.isDirectory() ) // если папка, то проверяем ее содержимое
			{
				result.addAll( compiledFiles( tmpdirFile ) );
			}
		}
		
		return result;
	}
	
	public static File getClassFile( String className )
    {
        // разделитель расширения ресурса (разбиваем на настоящее имя+расширение)
        String extension = getResourceExt(className);
        className = getResourceName(className);
        
        // разделитель для случая вложенного класса
        int innerClassDelim = className.indexOf( '$' );
        
        String path;
        if( innerClassDelim >= 0)
        {
            path = className.substring( 0, innerClassDelim );
        }
        else
        {
            path = className;
        }
        path = path.replace( '.', File.separatorChar );
        
        return new File( "dyn", path + "." + extension );
    }
	
	public static String getResourceExt(String className)
    {
        int resExtDelim = className.indexOf( ':' );
        return resExtDelim >= 0 ? className.substring( resExtDelim + 1 ).replace(":", ".") : "java";
    }
	
	public static String getResourceName(String className)
    {
        int resExtDelim = className.indexOf(':');
        return resExtDelim >= 0 ? className.substring( 0, resExtDelim ) : className;
    }

	/*
	 * удаляет директорию outputdir, в которой находятся скомпилированные файлы
	 */
	public void deleteClassDir()
	{
		if( outputdir != null && outputdir.exists() )
		{
			deleteDirWithFiles( outputdir );
		}
	}
	
	private void deleteDirWithFiles( File dir )
	{
		for( File tmpdirFile : dir.listFiles() )
		{
			if( tmpdirFile.isFile() )
			{
				tmpdirFile.delete();
			}
			else if( tmpdirFile.isDirectory() )
			{
				deleteDirWithFiles( tmpdirFile );
			}
		}
		
		dir.delete();
	}
	
	private class DiagnosticListenerImpl implements DiagnosticListener<JavaFileObject>
	{
		private CompilationResult result = new CompilationResult();
		
		private CompilationResult getCompilationResult()
		{
			return result;
		}
		
		@Override
		public void report( Diagnostic<? extends JavaFileObject> diagnostic )
		{
			CompilationMessage message = new CompilationMessage();
			message.setMessage( diagnostic.getMessage( Locale.ENGLISH ) );
			message.setLine( diagnostic.getLineNumber() );
			message.setColumn( diagnostic.getColumnNumber() );
			message.setSource( diagnostic.getSource() != null ? diagnostic.getSource().getName() : null );
			
			switch( diagnostic.getKind() )
			{
				case ERROR:
					result.getErrors().add( message );
					break;
				case WARNING:
				case MANDATORY_WARNING:
					result.getWarnings().add( message );
					break;
			}

			if( logger.isDebugEnabled() )
			{
				logger.debug(String.format("%s (%s) %s : %s [%d, %d]",
						diagnostic.getKind(), diagnostic.getCode(), diagnostic.getMessage( Locale.ENGLISH ),
						diagnostic.getSource() != null  ? diagnostic.getSource().getName() : null,
						diagnostic.getLineNumber(),  diagnostic.getColumnNumber()
						));
			}
		}
	}
	
	public static class CompiledUnit
	{
		String className;
		File classFile;
		File srcFile;
	}
	
//	@SuppressWarnings({ "restriction", "unchecked" })    
//    public static URL[] getUrls(ClassLoader classLoader) {
//        if (classLoader instanceof URLClassLoader) {
//            return ((URLClassLoader) classLoader).getURLs();
//        }
//
//        // jdk9
//        if (classLoader.getClass().getName().startsWith("jdk.internal.loader.ClassLoaders$")) {
//            try {
//                Field field = Unsafe.class.getDeclaredField("theUnsafe");
//                field.setAccessible(true);
//                Unsafe unsafe = (Unsafe) field.get(null);
//
//                // jdk.internal.loader.ClassLoaders.AppClassLoader.ucp
//                Field ucpField = classLoader.getClass().getDeclaredField("ucp");
//                long ucpFieldOffset = unsafe.objectFieldOffset(ucpField);
//                Object ucpObject = unsafe.getObject(classLoader, ucpFieldOffset);
//
//                // jdk.internal.loader.URLClassPath.path
//                Field pathField = ucpField.getType().getDeclaredField("path");
//                long pathFieldOffset = unsafe.objectFieldOffset(pathField);
//                ArrayList<URL> path = (ArrayList<URL>) unsafe.getObject(ucpObject, pathFieldOffset);
//
//                return path.toArray(new URL[path.size()]);
//            } catch (Exception e) {
//                e.printStackTrace();
//                return null;
//            }
//        }
//        return null;
//    }
}
