/*
 * Copyright (c) 2016 BiTel Inc. All rights reserved.
*/
package ru.bitel.mybgbilling.kernel.contract;

import java.io.Serializable;
import java.math.BigDecimal;
import java.time.ZoneId;
import java.time.temporal.TemporalAdjusters;
import java.util.Date;
import java.util.List;
import java.util.concurrent.TimeUnit;
import java.util.function.Supplier;

import javax.enterprise.context.SessionScoped;
import javax.inject.Inject;
import javax.inject.Named;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import ru.bitel.bgbilling.common.BGException;
import ru.bitel.bgbilling.common.BGRuntimeException;
import ru.bitel.bgbilling.kernel.contract.api.common.bean.Contract;
import ru.bitel.bgbilling.kernel.contract.limit.common.bean.ContractLimitMovement;
import ru.bitel.bgbilling.kernel.contract.limit.common.bean.ContractLimitParameters;
import ru.bitel.bgbilling.kernel.contract.limit.common.service.ContractLimitService;
import ru.bitel.common.function.Expirable;
import ru.bitel.common.function.Lazy;
import ru.bitel.mybgbilling.kernel.common.AbstractBean;
import ru.bitel.mybgbilling.kernel.common.inject.BGInject;
import ru.bitel.mybgbilling.kernel.common.inject.BGInjection;

@Named
@SessionScoped
@BGInjection
public class LimitBean
	extends AbstractBean
	implements Serializable
{
	private static final Logger logger = LogManager.getLogger();

	@BGInject(module = false)
	private ContractLimitService contractLimitService;
	
	@Inject
	private BalanceBean balanceBean;

	private Date month;

	private Supplier<BigDecimal> contractLimit;

	private Supplier<List<ContractLimitMovement>> contractLimitMovementList;

	private Lazy<ContractLimitParameters> contractLimitParameters;

	private int days;

	private BigDecimal amount;

	@Override
	protected void init()
		throws BGException
	{
		logger.info( "LimitBean.init()" );

		super.init();

		month = new Date();

		populate();
	}

	public Date getMonth()
	{
		return month;
	}

	public void setMonth( Date month )
	{
		if( month == null )
		{
			month = new Date();
		}

		this.month = month;
	}

	public int getDays()
	{
		return days;
	}

	public void setDays( int days )
	{
		this.days = days;
	}

	public BigDecimal getAmount()
	{
		return amount;
	}

	public void setAmount( BigDecimal amount )
	{
		if( amount != null )
		{
			amount = amount.abs();
		}

		this.amount = amount;
	}

	public BigDecimal getContractLimit()
		throws BGException
	{
		return contractLimit.get();
	}

	public List<ContractLimitMovement> getContractLimitMovementList()
		throws BGException
	{
		return contractLimitMovementList.get();
	}

	public ContractLimitParameters getContractLimitParameters()
		throws BGException
	{
		return contractLimitParameters.get();
	}

	public void populate()
		throws BGException
	{
		contractLimit = Expirable.asyncOf( () -> contractLimitService.contractLimitGet( getContractId() ), 120, TimeUnit.SECONDS );

		contractLimitMovementList = Expirable.of( () -> reverse( contractLimitService.contractLimitMovementList( getContractId(),
																												 Date.from( month.toInstant().atZone( ZoneId.systemDefault() )
																																 .toLocalDate()
																																 .withDayOfMonth( 1 )
																																 .atStartOfDay( ZoneId.systemDefault() )
																																 .toInstant() ),
																												 Date.from( month.toInstant().atZone( ZoneId.systemDefault() )
																																 .toLocalDate()
																																 .with( TemporalAdjusters.lastDayOfMonth() )
																																 .atStartOfDay( ZoneId.systemDefault() )
																																 .toInstant() ) ) ),
												  120, TimeUnit.SECONDS );

		contractLimitParameters = Lazy.of( () -> contractLimitService.contractLimitParametersGet( getContractId() ) );
	}

	public void beginLowerLimit()
		throws BGException
	{
		logger.debug( "beginLowerLimit" );

		populate();
		
		if( customerBean.getContract().getBalanceMode() != Contract.DEBET_BALANCE_MODE )
		{
			addErrorMessage( null, "Возможность понижения лимита заблокирована" );
			return;
		}

		try
		{
			contractLimitParameters.get();
		}
		catch( BGRuntimeException ex )
		{
			processException( logger, null, ex );
			return;
		}

		setStep( 1 );
	}

	public void lowerLimit()
		throws BGException
	{
		logger.debug( "limit" );

		if( amount == null || amount.compareTo( BigDecimal.ZERO ) <= 0 )
		{
			addErrorMessage( null, "Введите корректное значение суммы пополнения." );
			return;
		}

		contractLimitService.contractLimitMove( getContractId(), amount, days );

		amount = null;
		days = 0;

		setStep( 0 );
		
		try
		{
			balanceBean.populateCurrentContractBalance();
		}
		finally
		{
			populate();

			addInfoMessage( null, "Лимит был успешно понижен." );
		}
	}
}
