/*
 * Copyright (c) 2016 BiTel Inc. All rights reserved.
*/
package ru.bitel.mybgbilling.kernel.contract;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

import javax.enterprise.context.SessionScoped;
import javax.enterprise.inject.Instance;
import javax.inject.Inject;
import javax.inject.Named;

import ru.bitel.bgbilling.common.BGException;
import ru.bitel.bgbilling.common.BGMessageException;
import ru.bitel.bgbilling.kernel.tariff.option.common.bean.ContractTariffOption;
import ru.bitel.bgbilling.kernel.tariff.option.common.bean.ContractTariffOptionData;
import ru.bitel.bgbilling.kernel.tariff.option.common.bean.TariffOption;
import ru.bitel.bgbilling.kernel.tariff.option.common.service.TariffOptionService;
import ru.bitel.common.Utils;
import ru.bitel.common.function.Async;
import ru.bitel.common.function.Lazy;
import ru.bitel.mybgbilling.kernel.common.AbstractBean;
import ru.bitel.mybgbilling.kernel.common.Configuration;
import ru.bitel.mybgbilling.kernel.common.inject.BGInject;
import ru.bitel.mybgbilling.kernel.common.inject.BGInjection;
import ru.bitel.mybgbilling.plugins.bonus.BonusBean;

@SuppressWarnings("serial")
@Named
@SessionScoped
@BGInjection
public class TariffOptionBean	extends AbstractBean
{

	@BGInject( module = false )
	private TariffOptionService tariffOptionService;
	
	@Inject
	private BalanceBean balanceBean;
	
	@Inject
    private Configuration configuration;
	
	@Inject
	private Instance<BonusBean> bonusBeanInstance;


	private Async<List<ContractTariffOption>> contractTariffOptionList;
	private Lazy<List<ContractTariffOption>> historyContractTariffOptionList;

	private Lazy<List<TariffOption>> tariffOptionList;
	private Async<List<TariffOption>> availableTariffOptionList;
	private Lazy<List<TariffOption>> availableTariffOptionListFilter;
    private String availableTariffFilter;
	
	private Boolean isBonusPluginEnabled = null; // установлен ли и включен плагин
	private BigDecimal bonusBalance = null;
	private Map<Integer, String> modeBonusMap;
	private Set<Integer> optionIdOnlyBonuses = new HashSet<>();

	@Override
	protected void init()  throws BGException
	{
		super.init();
		populate();
	}

	public void populate() throws BGException
	{
		Date now = new Date();

		this.contractTariffOptionList = Async.of( ( ) -> tariffOptionService.contractTariffOptionListWeb( getContractId(), now, null ) );
		this.historyContractTariffOptionList = Lazy.of( ( ) -> {
			List<ContractTariffOption> history = tariffOptionService.contractTariffOptionHistory( getContractId(), now, true );
			if( history != null )
			{
				Collections.reverse( history );
			}
			return history;
		} );

		this.tariffOptionList = Lazy.of( ( ) -> Utils.maskNull( tariffOptionService.tariffOptionListAvailable( getContractId(), now, null, false, true ) ) );

		this.availableTariffOptionList = Async.of( ( ) -> {
			List<TariffOption> availableTariffOptionList = new ArrayList<TariffOption>();

			for( TariffOption o : this.tariffOptionList.get() )
			{
				if( o.isEnable() )
				{
					availableTariffOptionList.add( o );
				}
			}

			return availableTariffOptionList;
		} );
		
		if( isBonusPluginEnabled() )
        {
            optionIdOnlyBonuses = bonusBeanInstance.get().getOptionIdPayOnlyBonuses();
            
            if( bonusBeanInstance.get().isBonusPluginEnebledOfContract() )
            {
                bonusBalance = bonusBeanInstance.get().getBonusBalance();
                modeBonusMap = bonusBeanInstance.get().getModeBonusMap( availableTariffOptionList.get() );
            }
        }
		setAvailableTariffFilter(null);
		
	}

	public List<TariffOption> getTariffOptionList()    throws BGException
	{
		return tariffOptionList.get();
	}

	public List<TariffOption> getAvailableTariffOptionList()   throws BGException
	{
		//return availableTariffOptionList.get();
	    return availableTariffOptionListFilter.get();
	}

	public List<ContractTariffOption> getContractTariffOptionList()    throws BGException
	{
		return contractTariffOptionList.get();
	}

	public List<ContractTariffOption> getHistoryContractTariffOptionList() throws BGException
	{
		return historyContractTariffOptionList.get();
	}

	public void activate( int tariffOptionId, int tariffOptionActivateModeId, boolean payWithBonuses ) throws BGException
	{
	    boolean payOnlyBonus = isBonusPluginEnabled() && bonusBeanInstance.get().getOptionIdPayOnlyBonuses().contains( tariffOptionId );
	    if( payOnlyBonus && payWithBonuses == false )
        {
            throw new BGMessageException( "Активация данной тарифной опции возможна только бонусами!" );
        }
		try
		{
		    if( !payWithBonuses )
		    {
		        ContractTariffOptionData contractTariffOptionData = new ContractTariffOptionData();
		        contractTariffOptionData.setContractId( getContractId() );
		        contractTariffOptionData.setOptionId( tariffOptionId );
		        contractTariffOptionData.setModeId( tariffOptionActivateModeId );
		        contractTariffOptionData.setPeriod( null );
		        contractTariffOptionData.setWeb( true );
		        
		        tariffOptionService.contractTariffOptionActivate( contractTariffOptionData );
		        addInfoMessage( null, "Тарифная опция успешно активирована" );
		    }
		    else
		    {
		        bonusBeanInstance.get().activateTO( getContractId(), tariffOptionId, tariffOptionActivateModeId );
                addInfoMessage( null, "Тарифная опция с использованием бонусов успешно активирована" );
		    }
		}
		finally
		{
			try
			{
				balanceBean.populateCurrentContractBalance();
			}
			finally
			{
				populate();
			}
		}
	}

	public void deactivate( int contractTariffOptionId )   throws BGException
	{
		try
		{
			tariffOptionService.contractTariffOptionDeactivate( getContractId(), contractTariffOptionId );
			addInfoMessage( null, "Тарифная опция успешно деактивирована" );
		}
		finally
		{
			populate();
		}
	}

	public void reactivate( int contractTariffOptionId )   throws BGException
	{
		try
		{
			tariffOptionService.contractTariffOptionReactivate( getContractId(), contractTariffOptionId );
			addInfoMessage( null, "Тарифная опция успешно реактивирована" );
		}
		finally
		{
			populate();
		}
	}
	
	//////////////////////////////////////   Bonus   ////////////////////////////////////////
	/** Проверка факта ниличия плагина и его включенность в биллинге */
	private boolean isBonusPluginEnabled()
	{
	    if( isBonusPluginEnabled == null )
        {
            isBonusPluginEnabled = !bonusBeanInstance.isUnsatisfied() && configuration.isPluginEnabled( "bonus" );
        }
	    return isBonusPluginEnabled;
	}
	
	public boolean isCanPayBonuses() throws BGException
	{
	    return isBonusPluginEnabled() && bonusBeanInstance.get().isBonusPluginEnebledOfContract() && modeBonusMap != null && modeBonusMap.size() > 0 && bonusBalance != null && bonusBalance.compareTo( BigDecimal.ZERO ) == 1;
	}
	
	public Map<Integer, String> getModeBonusMap()
	{
	    return modeBonusMap;
	}
	
	public Set<Integer> getOptionIdOnlyBonuses()
	{
	    return optionIdOnlyBonuses;
	}

    public String getAvailableTariffFilter()
    {
        return availableTariffFilter;
    }

    public void setAvailableTariffFilter( String availableTariffFilter )
    {
        this.availableTariffFilter = availableTariffFilter;
        if(Utils.isEmptyString( availableTariffFilter) || availableTariffFilter.equals( "-1" ))
        {
            availableTariffOptionListFilter = Lazy.of( ()->availableTariffOptionList.get() );
        }
        else
        {
            Pattern p = Pattern.compile(availableTariffFilter);  
            availableTariffOptionListFilter = Lazy.of( ()->availableTariffOptionList.get().stream()
                                                       .filter( (i)->p.matcher( i.getTitle() ).matches() )
                                                       .collect(Collectors.toList()) );
            
        }
       
    }

    
}
