/*
 * Copyright (c) 2016 BiTel Inc. All rights reserved.
*/
package ru.bitel.mybgbilling.kernel.navigation;

import java.io.Serializable;
import java.math.BigDecimal;
import java.text.DateFormat;
import java.text.DecimalFormat;
import java.text.MessageFormat;
import java.text.NumberFormat;
import java.text.SimpleDateFormat;
import java.time.DayOfWeek;
import java.time.format.DateTimeFormatter;
import java.time.format.TextStyle;
import java.time.temporal.WeekFields;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.function.Function;

import javax.annotation.PostConstruct;
import javax.enterprise.context.SessionScoped;
import javax.faces.context.FacesContext;
import javax.inject.Inject;
import javax.inject.Named;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import ru.bitel.common.Utils;
import ru.bitel.common.model.IdTitle;
import ru.bitel.mybgbilling.kernel.common.Configuration;

@Named
@SessionScoped
public class LocaleBean
	implements Serializable
{
	private static final long serialVersionUID = 3820581315416670507L;

	private static final Logger logger = LogManager.getLogger();

	@Inject
	private Configuration configuration;

	private Locale locale;

	private ResourceBundle msg;

	private DateFormat dateFormat;

	private DateFormat monthFormat;
	private DateFormat monthFormat2;
	private DateFormat monthFormat3;
	
	private DateTimeFormatter dayOfWeekFormat;
	private DateTimeFormatter monthNameFormat;

	private NumberFormat numberFormat;

	private Function<String, BigDecimal> currencyParse;
	private Function<Object, String> currencyFormat;
	
	private List<IdTitle> week;

	@PostConstruct
	public void init()
	{
		FacesContext context = FacesContext.getCurrentInstance();

		locale = context.getExternalContext().getRequestLocale();

		initLocale( context );
	}

	private void initLocale( FacesContext context )
	{
		msg = context.getApplication().getResourceBundle( context, "msg" );

		monthFormat = new SimpleDateFormat( "LLLL yyyy", locale );
		monthFormat2 = new SimpleDateFormat( "MMMM yyyy", locale );
		monthFormat3 = new SimpleDateFormat( "mm.yyyy", locale );
		
		dayOfWeekFormat = DateTimeFormatter.ofPattern( "EEEE", locale );
		monthNameFormat = DateTimeFormatter.ofPattern( "LLLL", locale );

		// currency

		String key = "currency." + configuration.getCurrencyCode();
		if( msg.containsKey( key ) )
		{
			final String pattern = msg.getString( key );

			currencyFormat = ( a ) -> MessageFormat.format( pattern, a );

			final MessageFormat messageFormat = new MessageFormat( pattern );
			DecimalFormat decimalFormat = (DecimalFormat)messageFormat.getFormats()[0];
			decimalFormat.setParseBigDecimal( true );

			currencyParse = ( a ) -> {
				try
				{
					return (BigDecimal)messageFormat.parse( a )[0];
				}
				catch( Exception e )
				{
					return Utils.parseBigDecimal( a.replace( ',', '.' ), BigDecimal.ZERO );
				}
			};
		}
		else
		{
			NumberFormat format = NumberFormat.getCurrencyInstance( getLocale() );
			format.setCurrency( configuration.getCurrency() );
			format.setMinimumFractionDigits( 2 );
			((DecimalFormat)format).setParseBigDecimal( true );

			currencyFormat = ( a ) -> format.format( a );

			currencyParse = ( a ) -> {
				try
				{
					return (BigDecimal)format.parseObject( a );
				}
				catch( Exception e )
				{
					return Utils.parseBigDecimal( a.replace( ',', '.' ), BigDecimal.ZERO );
				}
			};
		}
		
		week = new ArrayList<IdTitle>();

		DayOfWeek day = WeekFields.of( getLocale() ).getFirstDayOfWeek();
		String title = day.getDisplayName( TextStyle.FULL, getLocale() );

		week.add( new IdTitle( day.getValue() == 7 ? 1 : day.getValue() + 1, title ) );

		for( int i = 0; i < 6; i++ )
		{
			day = day.plus( 1 );
			title = day.getDisplayName( TextStyle.FULL, getLocale() );

			week.add( new IdTitle( day.getValue() == 7 ? 1 : day.getValue() + 1, title ) );
		}

		numberFormat = NumberFormat.getNumberInstance( getLocale() );
	}

	public Locale getLocale()
	{
		return locale;
	}

	public String getLanguage()
	{
		return locale.getLanguage();
	}

	public void setLanguage( String language )
	{
		locale = new Locale( language );

		FacesContext context = FacesContext.getCurrentInstance();

		context.getViewRoot().setLocale( locale );

		initLocale( context );
	}

	public DateFormat getDateFormat()
	{
		return dateFormat;
	}

	public DateFormat getMonthFormat()
	{
		return monthFormat;
	}

	public DateFormat getMonthFormat2()
	{
		return monthFormat2;
	}

	public DateFormat getMonthFormat3()
	{
		return monthFormat3;
	}

	public String msg( String message )
	{
		if( msg.containsKey( message ) )
		{
			return msg.getString( message );
		}

		return message;
	}

	public String msg( String key, String def )
	{
		if( msg.containsKey( key ) )
		{
			return msg.getString( key );
		}

		if( msg.containsKey( def ) )
		{
			return msg.getString( def );
		}

		return def;
	}

	public String getFormat( String key )
	{
		if( msg.containsKey( key ) )
		{
			return msg.getString( key );
		}

		return null;
	}

	public String format( String key, Object arg )
	{
		if( msg.containsKey( key ) )
		{
			String pattern = msg.getString( key );

			return new MessageFormat( pattern, locale ).format( new Object[] { arg } );
		}

		return key;
	}

	public String format( String key, Object arg1, Object arg2 )
	{
		if( msg.containsKey( key ) )
		{
			String pattern = msg.getString( key );

			return new MessageFormat( pattern, locale ).format( new Object[] { arg1, arg2 } );
		}

		return key;
	}

	public String format( String key, Object arg1, Object arg2, Object arg3 )
	{
		if( msg.containsKey( key ) )
		{
			String pattern = msg.getString( key );

			return new MessageFormat( pattern, locale ).format( new Object[] { arg1, arg2, arg3 } );
		}

		return key;
	}

	public String format( String key, Object arg1, Object arg2, Object arg3, Object arg4 )
	{
		if( msg.containsKey( key ) )
		{
			String pattern = msg.getString( key );

			return new MessageFormat( pattern, locale ).format( new Object[] { arg1, arg2, arg3, arg4 } );
		}

		return key;
	}

	public String formatCurrency( Object value )
	{
		return currencyFormat.apply( value );
	}

	public String formatShortCurrency( Object value )
	{
		return numberFormat.format( value );
	}

	public BigDecimal parseCurrency( String value )
	{
		return currencyParse.apply( value );
	}

	public List<IdTitle> getWeek()
	{
        return week;
    }

    public DateTimeFormatter getDayOfWeekFormat()
    {
        return dayOfWeekFormat;
    }

    public static void main( String[] args )
    {
        System.out.println( MessageFormat.format( "{0,number,#,##0}", 123123123456789L ) );
	}

    public DateTimeFormatter getMonthNameFormat()
    {
        return monthNameFormat;
    }
}
