/*
 * Copyright (c) 2016 BiTel Inc. All rights reserved.
*/
package ru.bitel.mybgbilling.kernel.payment;

import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;

import javax.enterprise.context.ApplicationScoped;
import javax.enterprise.inject.Produces;
import javax.faces.context.ExternalContext;
import javax.faces.context.FacesContext;
import javax.inject.Qualifier;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import groovy.json.JsonOutput;
import ru.bitel.bgbilling.kernel.module.common.bean.BGModule;
import ru.bitel.bgbilling.kernel.module.common.service.ModuleService;
import ru.bitel.common.Utils;
import ru.bitel.common.groovy.GroovyBeanBuilder;
import ru.bitel.mybgbilling.kernel.common.inject.BGInject;
import ru.bitel.mybgbilling.kernel.common.inject.BGInjection;
import ru.bitel.mybgbilling.kernel.payment.PaymentConfig.PaymentItem;
import ru.bitel.mybgbilling.kernel.payment.PaymentConfig.PaymentProvider;

@BGInjection
public class PaymentConfigProducer
{
    @Qualifier
    @Retention( RetentionPolicy.RUNTIME )
    @Target( { ElementType.METHOD, ElementType.FIELD, ElementType.PARAMETER, ElementType.TYPE } )
    public @interface PaymentConfigProduce
    {
    }
    
	private static final Logger logger = LogManager.getLogger();

	@BGInject( module = false )
	private ModuleService moduleService;
	
    private static final Set<String> PAYMENT_MODULES = new HashSet<String>( Arrays.asList( "card",
        "agr", "assist", "alfabank", "bvcom", "mobimoney", "moneta",
        "mtsbank", "paymaster", "paylinks", "payture", "qiwi", "rbkmoney", "rfiec", "psb", "tinkoff",
        "pscb", "robokassa", "sberbank", "webmoney", "yamoney", "yoomoney", "payonline", "uniteller", "vepay" ) );

	@Produces
	@ApplicationScoped
	@PaymentConfigProduce
	public PaymentConfig newInstance()
	{
	    PaymentConfig result = null;

	    // сначала грузим данные из файла конфига
		try
		{
			result = new PaymentConfig();

			GroovyBeanBuilder bb = new GroovyBeanBuilder();
			bb.setClassLoader( this.getClass().getClassLoader() );
			bb.setInitialVariable( "paymentConfig", result );

			bb.setMappedClass( "payment", PaymentItem.class );
			bb.setMappedClass( "provider", PaymentProvider.class );

			ExternalContext externalContext = FacesContext.getCurrentInstance().getExternalContext();
			bb.load( externalContext.getResourceAsStream( "/WEB-INF/mybgbilling-payment.groovy" ), "mybgbilling-payment.groovy" );
		}
		catch( Exception e )
		{
			logger.error( "Error while parsing mybgbilling-payment.groovy" );
			logger.error( e.getMessage(), e );
		}

		if( result == null )
		{
			result = new PaymentConfig();
		}
		
		try
		{
			if( result.getProviders() == null )
			{
				result.setProviders( new ArrayList<PaymentProvider>() );
			}

			Map<Integer, PaymentProvider> providerMap =
				result.getProviders().stream().collect( Collectors.toMap( PaymentProvider::getModuleId, Function.identity(), ( a, b ) -> b ) );

			final List<BGModule> moduleList = moduleService.moduleList();

			// если не стоит флаг брать провайдеров только из конфига, то дополняем провайдеров из модулей  
			if( !result.isReplaceProviders() )
			{
				loadSystemProviders( result, providerMap, moduleList );
			}

			// сортируем провайдеров по названию (или ID, если названия нет)
            result.getProviders().sort( ( a, b ) -> Utils.maskBlank( a.getTitle(), Utils.maskNull( a.getId() ) )
                .compareTo( Utils.maskBlank( b.getTitle(), Utils.maskNull( b.getId() ) ) ) );

			// нормализуем конфиг провайдера
			try
			{
				result.getProviders().stream().filter( a -> a.getConfig() != null ).forEach( a -> a.getConfig().putAll( a.getConfig().flatten() ) );
			}
			catch( Exception ex )
			{
				logger.error( ex.getMessage(), ex );
			}

			// нормализуем конфиг режима платежа
			try
			{
				result.getPayments().stream().filter( a -> a.getConfig() != null ).forEach( a -> a.getConfig().putAll( a.getConfig().flatten() ) );
			}
			catch( Exception ex )
			{
				logger.error( ex.getMessage(), ex );
			}

			// если стоит флаг не показывать активацию карт - прячем модуль карт
			if( !result.isShowCard() )
			{
				result.getProviders().removeIf( ( a ) -> "card".equals( a.getModule() ) );
			}

			// проставляем провайдерам moduleId, если он не указан в конфиге
			result.getProviders().stream().filter( ( a ) -> a.getModuleId() <= 0 ).forEach( ( a ) -> {

				a.setModuleId( moduleList.stream()
				    .filter( ( b ) -> b.getName().equals( a.getModule() ) )
				    .findAny()
				    .map( ( c ) -> (Integer)c.getId() )
				    .orElse( 0 ) );
			} );

			// делаем уникальные ID
			result.getProviders().stream().collect( Collectors.toMap( PaymentProvider::getId, Function.<PaymentProvider> identity(), ( a, b ) -> {

				if( !a.getId().endsWith( String.valueOf( a.getModuleId() ) ) )
				{
					a.setId( a.getId() + a.getModuleId() );
				}

				b.setId( b.getId() + b.getModuleId() );

				return a;
			} ) );
			
			Map<String, PaymentProvider> providerIdMap =
				result.getProviders().stream().collect( Collectors.toMap( PaymentProvider::getId, Function.<PaymentProvider> identity(), ( a, b ) -> a ) );
			
			// проставляем режимам платежа поля из провайдеров
			result.getPayments().stream().forEach( a -> {

				PaymentProvider p = providerIdMap.get( a.getProviderId() );
				if( p == null )
				{
					return;
				}

                if( Utils.isBlankString( a.getImage() ) )
                {
                    a.setImage( p.getImage() );
                }

                if( Utils.isBlankString( a.getTitle() ) )
                {
                    a.setTitle( p.getTitle() );
                }

                if( a.getConfig() == null )
                {
                    a.setConfig( p.getConfig() );
                }
			} );
		}
		catch( Exception e )
		{
			logger.error( e.getMessage(), e );
		}

		logger.info( "Payment configuration:\n" + JsonOutput.prettyPrint( JsonOutput.toJson( result ) ) );

		return result;
	}

    private void loadSystemProviders( PaymentConfig result, Map<Integer, PaymentProvider> providerMap, final List<BGModule> moduleList )
    {
        // добавляем модуль card, если он еще не был добавлен в конфиге
        for( BGModule module : moduleList )
        {
            if( !"card".equals( module.getName() )
                || providerMap.containsKey( module.getId() ) )
            {
                continue;
            }

            PaymentProvider provider = new PaymentProvider();
            provider.setId( module.getName() );
            provider.setModule( module.getName() );
            provider.setModuleId( module.getId() );
            provider.setTitle( module.getTitle() );

            result.getProviders().add( provider );
            providerMap.put( module.getId(), provider );
        }

        // добляем системных провайдеров (модули), если они еще не указаны в конфиге
        // если указаны - то проставляем поля по умолчанию
        for( BGModule module : moduleList )
        {
        	if( !PAYMENT_MODULES.contains( module.getName() )  )
        	{
        		continue;
        	}
        	
        	final PaymentProvider exist = providerMap.get( module.getId() );

        	if( exist == null )
        	{
        		PaymentProvider provider = new PaymentProvider();
        		provider.setId( module.getName() );
        		provider.setModule( module.getName() );
        		provider.setModuleId( module.getId() );
        		provider.setTitle( module.getTitle() );

        		result.getProviders().add( provider );
        		providerMap.put( module.getId(), provider );
        	}
        	else
        	{
        		if( Utils.isBlankString( exist.getId() ) )
        		{
        			exist.setId( module.getName() );
        		}

        		if( Utils.isBlankString( exist.getModule() ) )
        		{
        			exist.setModule( module.getName() );
        		}

        		if( Utils.isBlankString( exist.getTitle() ) )
        		{
        			exist.setTitle( module.getTitle() );
        		}
        	}
        }
    }
}
