/*
 * Copyright (c) 2016 BiTel Inc. All rights reserved.
*/
package ru.bitel.mybgbilling.kernel.ws;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.Consumer;

import javax.xml.namespace.QName;
import javax.xml.soap.SOAPElement;
import javax.xml.soap.SOAPEnvelope;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPFactory;
import javax.xml.soap.SOAPHeader;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.Service;
import javax.xml.ws.handler.Handler;
import javax.xml.ws.handler.HandlerResolver;
import javax.xml.ws.handler.MessageContext;
import javax.xml.ws.handler.PortInfo;
import javax.xml.ws.handler.soap.SOAPHandler;
import javax.xml.ws.handler.soap.SOAPMessageContext;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import ru.bitel.bgbilling.kernel.container.ws.common.WSLinkedClient;
import ru.bitel.common.ref.ConcurrentReferenceHashMap;
import ru.bitel.common.ref.ConcurrentReferenceHashMap.ReferenceType;

public final class WSClient
	extends ru.bitel.bgbilling.kernel.container.ws.common.WSClient
	implements WSLinkedClient
{
	private static final Logger logger = LogManager.getLogger();

	private final String url;
	private final String username;
	private final String password;

	private final boolean logging;

	public WSClient( String url, String username, String password, boolean staticCache, boolean logging )
	{
		super( staticCache );

		this.url = url;
		this.username = username;
		this.password = password;

		this.logging = logging;
	}

	public <T> T getPort( Class<T> iface, int mid )
	{
		return getPort( url, iface, null, mid );
	}

	@Override
	public <T> T getPort( final String url, Class<T> iface, int mid )
	{
		return getPort( url, iface, null, mid );
	}

	@Override
	public HandlerResolver newHandlerResolver( final Service webService )
	{
		return new HandlerResolver()
		{
			@SuppressWarnings({ "rawtypes" })
			@Override
			public List<Handler> getHandlerChain( PortInfo portInfo )
			{
				List<Handler> result = new ArrayList<Handler>();
				result.add( new ServiceAuthenticationSOAPHandler( webService ) );
				return result;
			}
		};
	}

	class ServiceAuthenticationSOAPHandler
		implements SOAPHandler<SOAPMessageContext>
	{
		private final Service service;

		ServiceAuthenticationSOAPHandler( Service service )
		{
			this.service = service;
		}

		@Override
		public Set<QName> getHeaders()
		{
			return null;
		}

		@Override
		public void close( MessageContext context )
		{}

		@Override
		public boolean handleFault( SOAPMessageContext context )
		{
			if( context.getMessage().countAttachments() == 0 )
			{
				try
				{
					ByteArrayOutputStream out = new ByteArrayOutputStream( 500 );
					context.getMessage().writeTo( out );
					logger.debug( new String( out.toByteArray(), "UTF-8" ) );
				}
				catch( SOAPException e )
				{
					logger.error( e.getMessage(), e );
				}
				catch( IOException e )
				{
					logger.error( e.getMessage(), e );
				}
			}

			return true;
		}

		@Override
		public boolean handleMessage( SOAPMessageContext context )
		{
			try
			{
				boolean outbound = ((Boolean)context.get( SOAPMessageContext.MESSAGE_OUTBOUND_PROPERTY )).booleanValue();

				if( outbound )
				{
					QName operation = (QName)context.get( SOAPMessageContext.WSDL_OPERATION );

					if( logging )
					{
						logger.debug( service.getWSDLDocumentLocation() + " -> " + context.get( SOAPMessageContext.WSDL_SERVICE ) + (operation != null ? (":" + operation.getLocalPart()) : "") );
					}

					SOAPEnvelope se = context.getMessage().getSOAPPart().getEnvelope();
					SOAPHeader sh = se.getHeader();
					if( sh == null )
					{
						sh = se.addHeader();
					}

					SOAPFactory soapFactory = SOAPFactory.newInstance();

					// create SOAP elements specifying prefix and URI
					SOAPElement auth = soapFactory.createElement( "auth", "", "http://" + "ws" + ".base" + ".kernel.bgbilling.bitel.ru/" );

					auth.setAttribute( "user", username );
					auth.setAttribute( "pswd", password );

					sh.addChildElement( auth );
				}
				else
				{
					if( logging && context.getMessage().countAttachments() == 0 )
					{
						//context.getMessage().writeTo( System.err );
						//System.err.println();

						ByteArrayOutputStream out = new ByteArrayOutputStream( 500 );
						context.getMessage().writeTo( out );
						logger.debug( new String( out.toByteArray(), "UTF-8" ) );
					}
				}
			}
			catch( SOAPException e )
			{
				logger.error( e.getMessage(), e );
			}
			catch( IOException e )
			{
				logger.error( e.getMessage(), e );
			}

			return true;
		}
	}
	
	private ConcurrentReferenceHashMap<Class<?>, Consumer<Object>> gzipSupport = new ConcurrentReferenceHashMap<>( ReferenceType.WEAK, ReferenceType.SOFT );

	@SuppressWarnings("unchecked")
	@Override
	protected <T> T getPort( final String url, final Class<T> iface, final String module, final int mid )
	{
		final T port = super.getPort( url, iface, module, mid );

		final Map<String, Object> requestContext = ((BindingProvider)port).getRequestContext();

		try
		{
			// костыль, чтобы работала поддержка gzip в HTTP
			Consumer<Object> gzipSupport = this.gzipSupport.computeIfAbsent( port.getClass(), ( p ) -> {
				try
				{
					final Method m = p.getDeclaredMethod( "getInInterceptors" );
					final Class<?> clazz = m.getDeclaringClass().getClassLoader().loadClass( "org.apache.cxf.transport.common.gzip.GZIPInInterceptor" );

					return ( a ) -> {
						try
						{
							final Object o = clazz.newInstance();
							((List<Object>)m.invoke( a )).add( o );
						}
						catch( Throwable e )
						{
							logger.error( e.getMessage(), e );
						}
					};
				}
				catch( NoSuchMethodException e )
                {}
				catch( Throwable e )
				{
					logger.error( e.getMessage(), e );
				}

				return ( a ) -> {};
			} );

			gzipSupport.accept( port );
		}
		catch( Throwable e )
		{
			logger.error( e.getMessage(), e );
		}

		requestContext.put( "org.apache.cxf.transport.common.gzip.GZIPOutInterceptor.useGzip", "YES" );
		requestContext.put( "org.apache.cxf.transport.common.gzip.GZIPInInterceptor.useGzip", "YES" );

		return port;
	}
}
