/*
 * Copyright (c) 2016 BiTel Inc. All rights reserved.
*/
package ru.bitel.mybgbilling.modules.bill;

import java.math.BigDecimal;
import java.time.LocalDate;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.function.Supplier;
import java.util.stream.Collectors;

import javax.inject.Inject;
import javax.inject.Named;

import ru.bitel.bgbilling.common.BGException;
import ru.bitel.bgbilling.modules.bill.common.bean.BankAccount;
import ru.bitel.bgbilling.modules.bill.common.bean.DocType;
import ru.bitel.bgbilling.modules.bill.common.bean.DocumentPosition;
import ru.bitel.bgbilling.modules.bill.common.service.BillService;
import ru.bitel.common.Utils;
import ru.bitel.common.function.Async;
import ru.bitel.common.function.Lazy;
import ru.bitel.mybgbilling.kernel.common.AbstractBean;
import ru.bitel.mybgbilling.kernel.common.inject.BGInject;
import ru.bitel.mybgbilling.kernel.common.inject.BGInjection;
import ru.bitel.mybgbilling.kernel.common.scope.SessionModuleScoped;
import ru.bitel.mybgbilling.kernel.navigation.NavigationBean;

@Named
@SessionModuleScoped
@BGInjection
public class BillCreateBean
	extends AbstractBean
{
	@Inject
	private NavigationBean navigationBean;

	@Inject
	private BillBean billBean;

	@BGInject
	private BillService billService;

	private Async<List<DocType>> documentSpecList;
	private Async<List<BankAccount>> bankAccountList;

	private int documentSpecId;
	private int bankAccountId;

	private DocType documentSpec;
	private BankAccount bankAccount;

	private Supplier<List<DocumentPosition>> documentPositionList;

	@Override
	protected void init()
		throws BGException
	{
		super.init();

		populate();
	}

	public void populate()
		throws BGException
	{
		documentSpecList = Async.of( () -> Utils.maskNull( billService.documentSpecListForCustomer( getContractId() ) ) );
        bankAccountList = Async
            .of( () -> Utils.maskNull( billService.getBankAccountList( 0, true ) ).stream()
                .filter( a -> a.isInUse() )
                .collect( Collectors.toList() ) );

		setBankAccountId( 0 );
		step = 0;
	}

	public void populatePositionList()
		throws BGException
	{
		final LocalDate now = LocalDate.now();
		final int documentSpecId = this.documentSpecId;

		documentPositionList =
			Async.of( () -> Utils.maskNull( billService.templateDocumentPositionList( getContractId(), documentSpecId, now.getYear(), now.getMonthValue() - 1 ) ) );
	}

	public List<DocumentPosition> getDocumentPositionList()
	{
		return documentPositionList.get();
	}

	public List<DocType> getDocumentSpecList()
	{
		return documentSpecList.get();
	}

	public List<BankAccount> getBankAccountList()
	{
		return bankAccountList.get();
	}

	public int getDocumentSpecId()
	{
		return documentSpecId;
	}

	public void setDocumentSpecId( int documentSpecId )
		throws BGException
	{
		if( documentSpecId <= 0 )
		{
			this.step = 1;
			this.documentSpecId = 0;

			documentPositionList = Lazy.of( () -> Collections.emptyList() );

			return;
		}

		DocType documentSpec = documentSpecList.get().stream().filter( ( a ) -> a.getId() == documentSpecId ).findFirst().orElse( null );
		if( documentSpec == null )
		{
			this.step = 1;
			this.documentSpecId = 0;

			documentPositionList = Lazy.of( () -> Collections.emptyList() );

			return;
		}

		this.step = 2;

		this.documentSpec = documentSpec;
		this.documentSpecId = documentSpecId;

		populatePositionList();
	}

	public int getBankAccountId()
	{
		return bankAccountId;
	}

	public void setBankAccountId( int bankAccountId )
	{
		if( bankAccountList.get().size() == 1 )
		{
			this.step = 3;

			this.bankAccount = bankAccountList.get().get( 0 );
			this.bankAccountId = this.bankAccount.getId();

			return;
		}

		if( bankAccountId <= 0 )
		{
			this.step = 2;
			this.bankAccountId = 0;

			return;
		}

		BankAccount bankAccount = bankAccountList.get().stream().filter( ( a ) -> a.getId() == bankAccountId ).findFirst().orElse( null );
		if( bankAccount == null )
		{
			this.step = 2;
			this.bankAccountId = 0;

			return;
		}

		this.step = 3;

		this.bankAccount = bankAccount;
		this.bankAccountId = bankAccountId;
	}

	public DocType getDocumentSpec()
	{
		return documentSpec;
	}

	public BankAccount getBankAccount()
	{
		return bankAccount;
	}

	public void beginCreateBill()
		throws BGException
	{
		if( documentSpec == null )
		{
			addErrorMessage( null, "Выберите тип счёта." );
			return;
		}

		if( bankAccount == null )
		{
			addErrorMessage( null, "Выберите расчетный счет." );
			return;
		}

		// проверяем, что список есть, а не ошибка при получении такового;
		// чтобы ошибка вываливалась не на стадии рендеринга страницы
		documentPositionList.get();

		step = 4;

		navigationBean.setSubPage( "createBill" );
	}

	public void createBill()
		throws BGException
	{
		List<DocumentPosition> documentPositionList = new ArrayList<>( this.documentPositionList.get() );

		documentPositionList.removeIf( ( a ) -> a.getAmount() == null || BigDecimal.ZERO.compareTo( a.getAmount() ) == 0 );

		if( documentPositionList.size() == 0 )
		{
			addErrorMessage( null, "Укажите ненулевую сумму хотя бы в одной позиции." );
			return;
		}

		final LocalDate now = LocalDate.now();

		billService.billDocumentCreate( getContractId(), documentSpecId, bankAccountId, now.getYear(), now.getMonthValue() - 1, documentPositionList );

		setDocumentSpecId( 0 );

		step = 0;

		navigationBean.setSubPage( "" );

		billBean.populateBills();

		/*if( !conversation.isTransient() )
		{
			conversation.end();
		}*/
	}
}
