package ru.bitel.mybgbilling.modules.inet;

import java.time.LocalTime;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.concurrent.TimeUnit;
import java.util.function.Supplier;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.inject.Inject;
import javax.inject.Named;

import ru.bitel.bgbilling.common.BGException;
import ru.bitel.bgbilling.common.BGMessageException;
import ru.bitel.bgbilling.modules.inet.common.bean.InetServ;
import ru.bitel.bgbilling.modules.inet.common.bean.InetServRestriction;
import ru.bitel.bgbilling.modules.inet.common.bean.InetServRestrictionTimeCondition;
import ru.bitel.bgbilling.modules.inet.common.service.InetServService;
import ru.bitel.common.TimeUtils;
import ru.bitel.common.Utils;
import ru.bitel.common.function.Expirable;
import ru.bitel.mybgbilling.kernel.common.AbstractBean;
import ru.bitel.mybgbilling.kernel.common.inject.BGInject;
import ru.bitel.mybgbilling.kernel.common.inject.BGInjection;
import ru.bitel.mybgbilling.kernel.common.scope.ViewModuleScoped;

@Named
@ViewModuleScoped
@BGInjection
public class InetRestrictionBean
    extends AbstractBean
{
    private static final long serialVersionUID = -6888390075004902827L;

    @Inject
    private InetBean inetBean;

    @BGInject
    private InetServService inetServService;

    /**
     * Список временных ограничений.
     */
    private Supplier<List<InetServRestriction>> restrictions;

    private InetServRestriction restriction;

    @Override
    protected void init()
        throws BGException
    {
        populate();
    }

    private Stream<InetServRestriction> filterByTimeType( List<InetServRestriction> restrictions )
    {
        return Utils.maskNull( restrictions )
            .stream()
            .filter( r -> r.getType() == InetServRestriction.TYPE_TIME );
    }

    public void populate()
    {
        final int inetServId = inetBean.getInetServId();
        if( inetServId > 0 )
        {
            restrictions = Expirable.asyncOf(
                () -> filterByTimeType( inetServService.inetServRestrictionList( getContractId(), inetServId ) ).collect( Collectors.toList() ), 3, TimeUnit.MINUTES );
        }
        else
        {
            restrictions = Collections::emptyList;
        }
    }

    public List<InetServRestriction> getRestrictions()
    {
        return this.restrictions.get();
    }

    public InetServRestriction getRestriction()
    {
        return restriction;
    }

    public void editRestriction( int id )
        throws BGMessageException
    {
        InetServRestriction restriction = restrictions.get()
            .stream()
            .filter( r -> r.getId() == id )
            .findFirst()
            .orElse( null );

        if( restriction == null || restriction.getType() != InetServRestriction.TYPE_TIME )
        {
            throw new BGMessageException( "Ограничение не найдено." );
        }

        this.restriction = restriction;

        this.navigationBean.setSubPage( "edit" );
    }

    public void deleteRestriction( InetServRestriction restriction )
        throws BGException
    {
        InetServRestriction oldRestriction = filterByTimeType( restrictions.get() )
            .filter( r -> r.getId() == restriction.getId() )
            .findFirst()
            .orElse( null );

        if( oldRestriction == null || oldRestriction.getType() != InetServRestriction.TYPE_TIME )
        {
            throw new BGMessageException( "Ограничение не найдено." );
        }

        List<InetServRestriction> restrictions = Utils.maskNull( inetServService.inetServRestrictionList( getContractId(), restriction.getInetServId() ) );

        if( restrictions.removeIf( r -> r.getId() == restriction.getId() ) )
        {
            inetServService.inetServRestrictionListUpdate( getContractId(), restriction.getInetServId(), restrictions );
        }

        populate();
    }

    public void newTimeRestriction()
        throws BGMessageException
    {
        final InetServ inetServ = inetBean.getInetServ();
        if( inetServ == null || inetServ.getParentId() > 0 )
        {
            return;
        }

        if( restrictions.get().size() >= 5 )
        {
            throw new BGMessageException( "Создано максимальное количество правил." );
        }

        InetServRestriction restriction = new InetServRestriction();
        restriction.setInetServId( inetServ.getId() );
        restriction.setComment( "Создан абонентом" );

        InetServRestrictionTimeCondition timeCondition = new InetServRestrictionTimeCondition();
        timeCondition.setTimeFrom( LocalTime.of( 0, 0 ) );
        timeCondition.setTimeTo( LocalTime.of( 7, 0 ) );
        timeCondition.setDaysOfWeek( new HashSet<Integer>() );
        timeCondition.setDaysOfMonth( new HashSet<Integer>() );
        timeCondition.setMode( 0 );
        restriction.setConfig( timeCondition );

        this.restriction = restriction;

        this.navigationBean.setSubPage( "newSimple" );
    }

    public void saveSimpleRestriction()
        throws BGException
    {
        saveRestriction( true );
    }

    public void saveRestriction()
        throws BGException
    {
        saveRestriction( false );
    }

    public void saveRestriction( boolean simple )
        throws BGException
    {
        if( this.restriction == null )
        {
            return;
        }

        int inetServId = this.restriction.getInetServId();

        if( this.inetBean.getInetServList().stream().noneMatch( s -> s.getId() == inetServId ) )
        {
            throw new BGException( "Указанный аккаунт не найден." );
        }
        
        InetServRestrictionTimeCondition timeCondition = (InetServRestrictionTimeCondition)this.restriction.getConfig();
        if ( timeCondition.getTimeTo().isBefore( timeCondition.getTimeFrom() ) )
        {
            throw new BGException( "Время начала должно быть меньше времени окончания." );
        }

        List<InetServRestriction> restrictions = Utils.maskNull( inetServService.inetServRestrictionList( getContractId(), this.restriction.getInetServId() ) );

        if( this.restriction.getId() <= 0 )
        {
            if( filterByTimeType( restrictions ).count() >= 5 )
            {
                throw new BGMessageException( "Создано максимальное количество правил." );
            }

            if( this.restriction.getDateFrom() == null || TimeUtils.dateBefore( this.restriction.getDateFrom(), new Date() ) )
            {
                this.restriction.setDateFrom( new Date() );
            }

            if( simple )
            {
                timeCondition = (InetServRestrictionTimeCondition)restriction.getConfig();
                timeCondition.setDaysOfMonth( Collections.emptySet() );
                timeCondition.setMonths( Collections.emptySet() );

                this.restriction.setDateTo( null );
            }

            restrictions = new ArrayList<InetServRestriction>( restrictions );
            restrictions.add( this.restriction );
        }
        else
        {
            InetServRestriction oldRestriction = restrictions.stream().filter( r -> r.getId() == this.restriction.getId() ).findFirst().orElse( null );
            if( oldRestriction != null )
            {
                checkDatesOnUpdate( restrictions, oldRestriction );

                if( restrictions.removeIf( r -> r.getId() == this.restriction.getId() ) )
                {
                    restrictions.add( this.restriction );
                }
            }
        }

        inetServService.inetServRestrictionListUpdate( getContractId(), inetServId, restrictions );

        populate();
    }

    private void checkDatesOnUpdate( List<InetServRestriction> restrictions, InetServRestriction oldRestriction )
    {
        if( oldRestriction.getDateTo() != null
            && !TimeUtils.dateEqual( oldRestriction.getDateTo(), this.restriction.getDateTo() )
            && TimeUtils.dateBefore( this.restriction.getDateTo(), new Date() )
            && TimeUtils.dateBefore( restriction.getDateTo(), new Date() ) )
        {
            this.restriction.setDateTo( oldRestriction.getDateTo() );
        }

        if( restrictions.removeIf( r -> r.getId() == this.restriction.getId() ) )
        {
            restrictions.add( this.restriction );
        }

        if( oldRestriction.getDateFrom() != null
            && !TimeUtils.dateEqual( oldRestriction.getDateFrom(), this.restriction.getDateFrom() )
            && TimeUtils.dateBefore( this.restriction.getDateFrom(), new Date() )
            && TimeUtils.dateBefore( restriction.getDateFrom(), new Date() ) )
        {
            this.restriction.setDateFrom( oldRestriction.getDateFrom() );
        }
    }
}
