/*
 * Copyright (c) 2016 BiTel Inc. All rights reserved.
*/
package ru.bitel.mybgbilling.modules.trayinfo;

import java.io.Serializable;
import java.security.SecureRandom;
import java.time.LocalDate;
import java.time.ZoneId;
import java.util.Date;
import java.util.List;
import java.util.concurrent.TimeUnit;
import java.util.function.Supplier;
import java.util.stream.Collectors;

import javax.enterprise.context.SessionScoped;
import javax.inject.Named;

import ru.bitel.bgbilling.common.BGException;
import ru.bitel.bgbilling.modules.trayinfo.common.TrayInfoLogin;
import ru.bitel.bgbilling.modules.trayinfo.common.TrayInfoType;
import ru.bitel.bgbilling.modules.trayinfo.common.service.TrayInfoService;
import ru.bitel.common.TimeUtils;
import ru.bitel.common.Utils;
import ru.bitel.common.function.Expirable;
import ru.bitel.mybgbilling.kernel.common.AbstractBean;
import ru.bitel.mybgbilling.kernel.common.inject.BGInject;
import ru.bitel.mybgbilling.kernel.common.inject.BGInjection;

@Named
@SessionScoped
@BGInjection
public class TrayInfoBean
	extends AbstractBean
	implements Serializable
{
	private static final long serialVersionUID = 1540623437095557110L;

	@BGInject
	private TrayInfoService trayInfoService;

	private Supplier<List<TrayInfoType>> trayInfoTypeList;

	private Supplier<List<TrayInfoLogin>> trayInfoLoginList;

	private int trayInfoTypeId;

	private String password;

	@Override
	protected void init()
		throws BGException
	{
		super.init();

		password = generateSimplePassword( 6 );

		populate();
	}

	public void populate()
		throws BGException
	{
		trayInfoTypeList = Expirable.asyncOf( () -> Utils.maskNull( trayInfoService.getLoginTypes() ).stream()
														 .filter( ( a ) -> TimeUtils.dateInRange( new Date(), a.getDate1(), a.getDate2() ) )
														 .collect( Collectors.toList() ),
											  5, TimeUnit.MINUTES );

		Date showPeriodStart = Date.from( LocalDate.now().minus( java.time.Period.ofMonths( 2 ) ).atStartOfDay( ZoneId.systemDefault() ).toInstant() );

		trayInfoLoginList = Expirable.asyncOf( () -> Utils.maskNull( trayInfoService.getLogins( getContractId() ) ).stream()
														  .filter( ( a ) -> TimeUtils.checkDateIntervalsIntersection( showPeriodStart, new Date(), a.getDate1(), a.getDate2() ) )
														  .collect( Collectors.toList() ),
											   1, TimeUnit.MINUTES );
	}

	public List<TrayInfoType> getTrayInfoTypeList()
	{
		return trayInfoTypeList.get();
	}

	public List<TrayInfoLogin> getTrayInfoLoginList()
	{
		return trayInfoLoginList.get();
	}

	public int getTrayInfoTypeId()
	{
		return trayInfoTypeId;
	}

	public void setTrayInfoTypeId( int trayInfoTypeId )
	{
		this.trayInfoTypeId = trayInfoTypeId;
	}

	public String getPassword()
	{
		return password;
	}

	public void setPassword( String password )
	{
		this.password = password;
	}

	public void activate()
		throws BGException
	{
		if( getTrayInfoLoginList().stream().filter( ( a ) -> TimeUtils.dateInRange( new Date(), a.getDate1(), a.getDate2() ) ).count() > 0 )
		{
			addErrorMessage( null, "У вас уже активирован логин." );
			return;
		}

		if( trayInfoTypeId <= 0 || getTrayInfoTypeList().stream().filter( ( a ) -> a.getId() == trayInfoTypeId ).count() == 0 )
		{
			addErrorMessage( null, "Выберите режим активации." );
			return;
		}

		TrayInfoLogin login = new TrayInfoLogin();
		login.setContractId( getContractId() );
		login.setType( trayInfoTypeId );
		login.setPassword( Utils.maskNull( password ) );

		try
		{
			trayInfoService.updateLogin( login );

			setStep( 0 );

			password = generateSimplePassword( 6 );
		}
		finally
		{
			populate();
		}
	}

	private String generateSimplePassword( int length )
	{
		SecureRandom random = new SecureRandom();

		StringBuilder sb = new StringBuilder();

		for( int i = 0; i <= length; i++ )
		{
			sb.append( random.nextInt( 10 ) );
		}

		String passwordChars = sb.toString();
		int passwordCharsLength = passwordChars.length();

		sb.setLength( 0 );

		for( int i = 0; i < length; i++ )
		{
			sb.appendCodePoint( passwordChars.codePointAt( random.nextInt( passwordCharsLength ) ) );
		}

		return sb.toString();
	}
}
