/*
 * Copyright (c) 2016 BiTel Inc. All rights reserved.
*/
package ru.bitel.mybgbilling.modules.voiceip;

import java.io.Serializable;
import java.time.LocalDate;
import java.time.YearMonth;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.function.Supplier;
import java.util.stream.Collectors;

import javax.inject.Inject;
import javax.inject.Named;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import ru.bitel.bgbilling.common.BGException;
import ru.bitel.common.Utils;
import ru.bitel.common.function.Async;
import ru.bitel.common.function.Lazy;
import ru.bitel.common.model.IdTitle;
import ru.bitel.common.model.Page;
import ru.bitel.mybgbilling.kernel.common.AbstractBean;
import ru.bitel.mybgbilling.kernel.common.Account;
import ru.bitel.mybgbilling.kernel.common.MyApplicationBean.HttpRequest;
import ru.bitel.mybgbilling.kernel.common.inject.BGInjection;
import ru.bitel.mybgbilling.kernel.common.scope.SessionModuleScoped;
import ru.bitel.mybgbilling.kernel.common.utils.XmlUtils;

@Named
@SessionModuleScoped
@BGInjection
public class VoiceIpBean
	extends AbstractBean
	implements Serializable
{
	private static final long serialVersionUID = 1540623437095557110L;

	private static final Logger logger = LogManager.getLogger();

	@Inject
	private transient HttpRequest httpRequest;

	private Supplier<List<Account>> accountList;

	private int accountId;
	private IdTitle account;

	private Supplier<Set<Integer>> accountIds;

	private LocalDate dateFrom;
	private LocalDate dateTo;
	
	private int callType;

	public int getCallType()
	{
		return callType;
	}

	public void setCallType( int callType )
	{
		this.callType = callType;
	}

	public Set<Integer> getAccountIds()
	{
		return accountIds != null ? accountIds.get() : null;
	}

	public void setAccountIds( Set<Integer> loginIds )
	{
		if( loginIds != null )
		{
			loginIds.retainAll( getAccountList().stream().map( a -> a.getId() ).collect( Collectors.toSet() ) );
		}

		this.accountIds = () -> loginIds;
	}

	@Override
	protected void init()
		throws BGException
	{
		super.init();

		dateFrom = LocalDate.now().withDayOfMonth( 1 );
		dateTo = YearMonth.from( dateFrom ).atEndOfMonth();

		populate();

		this.accountIds = Async.of( () -> accountList.get().stream().map( a -> a.getId() ).collect( Collectors.toSet() ) );
	}

	public void populate()
		throws BGException
	{
		accountList = Async.of( () -> {

			List<Account> result = new ArrayList<>();

			try
			{
				Document doc = httpRequest.requestXml( "voiceip", getModuleId(), "GetLogins", getContractId() );

				DateTimeFormatter formatter = DateTimeFormatter.ofPattern( "dd.MM.yyyy" );

				for( Element e : XmlUtils.selectElements( doc, "//logins/item" ) )
				{
					int id = Utils.parseInt( e.getAttribute( "id" ) );
					if( id > 0 )
					{
						Account a = new Account();
						a.setId( id );
						a.setTitle( e.getAttribute( "title" ) );

						try
						{
							a.setDateFrom( formatter.parse( e.getAttribute( "date1" ), LocalDate::from ) );
						}
						catch( Exception ex )
						{}

						try
						{
							a.setDateTo( formatter.parse( e.getAttribute( "date2" ), LocalDate::from ) );
						}
						catch( Exception ex )
						{}

						result.add( a );
					}
				}

			}
			catch( Throwable ex )
			{
				logger.error( ex.getMessage(), ex );
			}
			return result;
		} );
	}

	public LocalDate getDateFrom()
	{
		return dateFrom;
	}

	public void setDateFrom( LocalDate dateFrom )
	{
		this.dateFrom = dateFrom;

		if( this.dateFrom == null )
		{
			this.dateFrom = LocalDate.now().withDayOfMonth( 1 );
		}
	}

	public LocalDate getDateTo()
	{
		return dateTo;
	}

	public void setDateTo( LocalDate dateTo )
	{
		this.dateTo = dateTo;

		if( this.dateTo == null || !YearMonth.from( this.dateFrom ).equals( YearMonth.from( this.dateTo ) ) )
		{
			this.dateTo = YearMonth.from( this.dateFrom ).atEndOfMonth();
		}
	}

	public int getAccountId()
	{
		return accountId;
	}

	public void setAccountId( int loginId )
		throws BGException
	{
		this.accountId = loginId;

		this.account = null;

		for( IdTitle login : getAccountList() )
		{
			if( login.getId() == loginId )
			{
				this.account = login;
				break;
			}
		}

		if( this.account == null )
		{
			this.accountId = 0;
		}
	}

	public IdTitle getAccount()
	{
		return account;
	}

	public List<Account> getAccountList()
	{
		return accountList.get();
	}

	Supplier<Page> getPage( Supplier<Document> documentSupplier, Page page )
	{
		return Lazy.of( () -> {

			Document doc = documentSupplier.get();
			if( doc == null )
			{
				return new Page( 1, 15 );
			}

			Page result = new Page( 1, 15 );

			result.setPageIndex( Utils.parseInt( XmlUtils.select( doc, "/data/table/@pageIndex" ), page.getPageIndex() ) );
			result.setPageSize( Utils.parseInt( XmlUtils.select( doc, "/data/table/@pageSize" ), page.getPageSize() ) );
			result.setPageCount( Utils.parseInt( XmlUtils.select( doc, "/data/table/@pageCount" ), page.getPageCount() ) );
			result.setRecordCount( Utils.parseInt( XmlUtils.select( doc, "/data/table/@recordCount" ), page.getRecordCount() ) );
			result.setPageIndex( Math.min( result.getPageIndex(), result.getPageCount() ) );

			return result;
		} );
	}
}
