/*
 * Copyright (c) 2016 BiTel Inc. All rights reserved.
*/
package ru.bitel.mybgbilling.plugins.dispatch;

import java.util.List;

import javax.enterprise.context.SessionScoped;
import javax.inject.Named;

import ru.bitel.bgbilling.common.BGException;
import ru.bitel.bgbilling.common.BGIllegalAccessException;
import ru.bitel.bgbilling.common.BGMessageException;
import ru.bitel.bgbilling.plugins.dispatch.common.service.DispatchService;
import ru.bitel.bgbilling.plugins.dispatch.common.bean.Contact;
import ru.bitel.bgbilling.plugins.dispatch.common.bean.ContactType;
import ru.bitel.bgbilling.plugins.dispatch.common.bean.Subscription;
import ru.bitel.common.Utils;
import ru.bitel.common.function.Async;
import ru.bitel.mybgbilling.kernel.common.AbstractConversationBean;
import ru.bitel.mybgbilling.kernel.common.inject.BGInject;
import ru.bitel.mybgbilling.kernel.common.inject.BGInjection;

/**
 * Плагин для работы с рассылками (dispatch).
 * 
 * @author marsel
 *
 */
@Named
@SessionScoped
@BGInjection
public class DispatchBean
	extends AbstractConversationBean
{
	@BGInject
	private DispatchService dispatchService;

	private Async<List<Contact>> contacts;
	private Async<List<ContactType>> contactTypes;
	private Async<List<Subscription>> subscriptions;
	
	private int contactTypeId;
	private String contactValue;
	
	@Override
	protected void init()
		throws BGException
	{
		super.init();

		populate();
		setStep( 0 );
	}

	public void populate()
		throws BGException
	{
		contactTypes = Async.of( () -> dispatchService.getContactTypes() );
		
		populateContacts();
		populateSubscriptions();
	}
	
	private void populateContacts()
	{
		contacts = Async.of( () -> dispatchService.getContacts( getContractId() ) );
	}
	
	public void populateSubscriptions()
	{
		subscriptions = Async.of( () -> dispatchService.getSubscriptions( 0, getContractId() ) );
	}
	
	public List<Contact> getContacts()
	{
		return contacts.get();
	}
	
	public List<Subscription> getSubscriptions() 
	{
		return subscriptions.get();
	}
	
	/**
	 * Получение рассылки по id
	 * @param subscriptionId - id
	 * @return
	 */
	public Subscription getSubscriptionById( int subscriptionId ) 
	{
		return subscriptions.get().stream()
			.filter( subscription -> subscription.getId() == subscriptionId )
			.findFirst()
			.orElse( null );
	}
	
	public List<ContactType> getContactTypes()
	{
		return contactTypes.get();
	}
	
	public int getContactTypeId()
	{
		return contactTypeId;
	}

	public void setContactTypeId( int contactTypeId )
	{
		this.contactTypeId = contactTypeId;
	}

	public String getContactValue()
	{
		return contactValue;
	}

	public void setContactValue( String contactValue )
	{
		this.contactValue = contactValue;
	}
	
	@Override
	public void setStep( int step ) {
		super.setStep( step );
		contactTypeId = -1;
		contactValue = "";
	}
	
	/**
	 * Получение типа контакта по id
	 * 
	 * @param contactTypeId
	 * @return
	 * @throws BGIllegalAccessException
	 */
	public ContactType getContactTypeById( String contactTypeId )
		throws BGIllegalAccessException
	{
		if ( Utils.isBlankString( contactTypeId ) )
		{
			throw new BGIllegalAccessException();
		}

		int id = Utils.parseInt( contactTypeId );

		return contactTypes.get().stream()
			.filter( contact -> contact.getId() == id )
			.findFirst()
			.orElse( new ContactType() );
	}
	
	/**
	 * Создание контакта
	 * 
	 * @throws BGException
	 */
	public void createContact()
		throws BGException
	{
		if ( Utils.isBlankString( contactValue ) ) 
		{
			throw new BGMessageException( "Укажите значение" );
		}
		
		Contact contact = new Contact();
		contact.setContractId( getContractId() );
		contact.setContactTypeId( contactTypeId );
		contact.setValue( contactValue );
		
		dispatchService.updateContact( contact );
		
		/**
		 * В случае, если создание контакта прошло успешно, то необходимо
		 * сбросить шаг на предыдущий (для отображения на странице кнопки
		 * создать)
		 */
		setStep( 0 );
		
		populateContacts();
	}
	
	/**
	 * Редактирование контакта
	 * 
	 * @param contact
	 *            - обновленный контакт, который необходимо сохранить
	 * @throws BGException
	 */
	public void updateContact( Contact contact ) 
		throws BGException 
	{
		/**
		 * Проверка на <= 0 нужна для того, чтобы при неверно переданном id по
		 * ошибке не создавался новый контакт. Т.к. данный метод предназначен
		 * для обновления, а не создания.
		 */
		if ( contact == null || contact.getId() <= 0 )
		{
			throw new BGIllegalAccessException();
		}
		
		/**
		 * Проверка нужна на тот случай, когда пользователь просто нажал
		 * "Редактировать" и сразу "Сохранить", оставив значение без изменений.
		 */
		if ( !contactValue.equals( "" ) ) 
		{
			contact.setValue( contactValue );
		}
		
		try 
		{
			dispatchService.updateContact( contact );
		}
		finally 
		{
			setStep( 0 );
			populateContacts();
			populateSubscriptions();
		}
	}
	
	/**
	 * Удаление контакта по id
	 * 
	 * @param contactId
	 * @throws BGException
	 */
	public void deleteContactById( String contactId )
		throws BGException
	{
		if ( Utils.isBlankString( contactId ) )
		{
			throw new BGIllegalAccessException();
		}

		int id = Utils.parseInt( contactId );

		dispatchService.deleteContact( id );
		
		populateContacts();
	}
	
	/**
	 * Удаление рассылки по id
	 * 
	 * @param subscriptionId
	 * @throws BGException
	 */
	public void deleteSubscriptionById( String subscriptionId )
		throws BGException
	{
		if ( Utils.isBlankString( subscriptionId ) )
		{
			throw new BGIllegalAccessException();
		}
	
		int id = Utils.parseInt( subscriptionId );
	
		dispatchService.deleteSubscription( id );
		
		populateSubscriptions();
	}
}
