/*
 * Copyright (c) 2016 BiTel Inc. All rights reserved.
*/
package ru.bitel.mybgbilling.plugins.dispatch;

import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

import javax.inject.Inject;
import javax.inject.Named;

import ru.bitel.bgbilling.common.BGException;
import ru.bitel.bgbilling.common.BGIllegalAccessException;
import ru.bitel.bgbilling.plugins.dispatch.common.service.DispatchService;
import ru.bitel.bgbilling.plugins.dispatch.common.bean.Contact;
import ru.bitel.bgbilling.plugins.dispatch.common.bean.Dispatch;
import ru.bitel.bgbilling.plugins.dispatch.common.bean.Subscription;
import ru.bitel.common.Preferences;
import ru.bitel.common.Utils;
import ru.bitel.common.function.Async;
import ru.bitel.common.model.Period;
import ru.bitel.common.model.ScheduleTime;
import ru.bitel.mybgbilling.kernel.common.AbstractConversationBean;
import ru.bitel.mybgbilling.kernel.common.inject.BGInject;
import ru.bitel.mybgbilling.kernel.common.inject.BGInjection;
import ru.bitel.mybgbilling.kernel.common.scope.SessionModuleScoped;

/**
 * Бин для подписки на рассылки
 * 
 * @author marsel
 *
 */
@Named
@SessionModuleScoped
@BGInjection
public class DispatchSubscribeBean
	extends AbstractConversationBean
{
	@BGInject
	private DispatchService dispatchService;
	
	@Inject
	private DispatchBean dispatchBean;

	private Async<List<Dispatch>> dispatches;
	private int dispatchId;
	private String comment;
	private boolean isEdit;
	private Subscription subscription;
	
	private Set<Integer> contactsIds;
	private Set<String> daysIds;
	
	@Override
	protected void init()
		throws BGException
	{
		super.init();

		populate();
	}

	public void populate()
		throws BGException
	{
		dispatches = Async.of( () -> dispatchService.getDispatchList() );
	}

	public List<Dispatch> getDispatches()
	{
		return dispatches.get();
	}
	
	public int getDispatchId()
	{
		return dispatchId;
	}

	public void setDispatchId( int dispatchId )
	{
		this.dispatchId = dispatchId;
	}
	
	public String getComment()
	{
		return comment;
	}

	public void setComment( String comment )
	{
		this.comment = comment;
	}
	
	public Set<Integer> getContactsIds()
	{
		return contactsIds;
	}

	public void setContactsIds( Set<Integer> contactIds )
	{
		this.contactsIds = contactIds;
	}
	
	public Set<String> getDaysIds()
	{
		return daysIds;
	}

	public void setDaysIds( Set<String> daysIds )
	{
		this.daysIds = daysIds;
	}
	
	/**
	 * Получение текущей рассылки
	 * @return
	 * @throws BGException
	 */
	public Dispatch getDispatch()
		throws BGException
	{
		return dispatches.get().stream()
			.filter( dispatch -> dispatch.getId() == dispatchId )
			.findFirst()
			.orElse( null );
	}
	
	/**
	 * Начать подписку 
	 */
	public void beginSubscribe()
	{
		comment = "";
		isEdit = false;
		
		navigationBean.setSubPage( "subscribeDispatch" );
	}
	
	/**
	 * Начать редактирование подписки
	 * @param id - id подписки
	 * @throws BGException
	 */
	public void beginEditSubscribtion( int id )
		throws BGException
	{
		if ( id < 0 )
		{
			throw new BGIllegalAccessException();
		}
		
		isEdit = true;
		subscription = dispatchBean.getSubscriptionById( id );
		dispatchId = subscription.getDispatch().getId();
		comment = subscription.getComment();
		contactsIds = new HashSet<Integer>();
		for ( Contact contact : subscription.getContacts() )
		{
			contactsIds.add( contact.getId() );
		}
		daysIds = convertScheduleTimeStringToSetIds( subscription );
			
		navigationBean.setSubPage( "subscribeDispatch" );
	}
	
	/**
	 * Подписаться на рассылку 
	 * @throws BGException
	 */
	public void subscribe() 
		throws BGException
	{
		if ( isEdit ) 
		{
			editSubscription();
		}
		else
		{
			createSubscription();
		}
		endSubscribe();
	}
	
	/**
	 * Создание новой подписки на рассылку
	 * @throws BGException
	 */
	private void createSubscription() 
		throws BGException
	{
		Subscription subscription = new Subscription();
		subscription.setContractId( getContractId() );
		
		List<Contact> contacts = dispatchBean.getContacts().stream()
			.filter( contact -> contactsIds.contains( String.valueOf( contact.getId() ) ) )
			.collect( Collectors.toList() );
		
		subscription.setContacts( contacts );
		if ( getDispatch().isPersonal() )
		{
			subscription.setRepeatTime( convertDaysToScheduleTime() );
		}
		// В БД колонка комментария не может быть null 
		subscription.setComment( ( comment != null ) ? comment : "" );
		subscription.setDispatch( getDispatch() );
		subscription.setActive( true );
		// TODO: В старом личном кабинете немного неясная логика, необходимо решить, что делать здесь
		subscription.setPreferences( new Preferences() );
		subscription.setSubscriptionPeriod( new Period() );
		
		dispatchService.updateSubscriptions( subscription );
	}
	
	/**
	 * Редактирование существующей подписки на рассылку
	 * @throws BGException
	 */
	private void editSubscription() 
		throws BGException
	{
		subscription.setComment( comment );
		
		List<Contact> contacts = dispatchBean.getContacts().stream()
			.filter( contact -> contactsIds.contains( String.valueOf( contact.getId() ) ) )
			.collect( Collectors.toList() );
	
		subscription.setContacts( contacts );
		subscription.setRepeatTime( convertDaysToScheduleTime() );
		
		dispatchService.updateSubscriptions( subscription );
	}
	
	public void endSubscribe() 
	{
		step = 0;
		navigationBean.setSubPage( "dispatches" );
		contactsIds = null;
		daysIds = null;

		dispatchBean.populateSubscriptions();

		if ( !conversation.isTransient() )
		{
			conversation.end();
		}
	}
	
	/**
	 * Получение списка контактов, которые соответствуют только типу контактов
	 * текущей рассылки
	 * 
	 * @return - список контактов
	 * @throws BGException 
	 */
	public List<Contact> getContactListByCurrentDispatch()
		throws BGException
	{
		Dispatch currentDispatch = dispatches.get().stream()
			.filter( dispatch -> dispatch.getId() == dispatchId )
			.findAny()
			.orElse( new Dispatch() );
		
		List<Contact> contacts = dispatchBean.getContacts();
		
		return contacts.stream()
			.filter( contact -> contact.getContactTypeId() == currentDispatch.getContactTypeId() )
			.collect( Collectors.toList() );
	}
	
	/**
	 * Преобразование набора id в ScheduleTime
	 * @return
	 */
	private ScheduleTime convertDaysToScheduleTime()
	{
		ScheduleTime time = new ScheduleTime();
		time.setMonth( 0 );
		time.setDayOfWeek( (int)Utils.packValues( daysIds, 1 ) );
		time.setDay( 0 );
		time.setHour( (int)Utils.packValues( Utils.toList( "0" ), 0 ) );
		time.setMin( (int)Utils.packValues( Utils.toList( "0" ), 0 ) );
		return time;
	}
	
	/**
	 * Преобразование строки расписания вида "1-2,5-7" в множество 1,2,5,6,7  
	 * @param subscription
	 * @return
	 */
	private Set<String> convertScheduleTimeStringToSetIds( Subscription subscription )
	{
		Set<String> result = new HashSet<>();
		
		String days = subscription.getRepeatTime().getDayOfWeekAsString();
		if ( !Utils.isEmptyString( days ) )
		{
			try {
				String[] daysIdsArray = days.split( "," );
				for ( int i = 0; i < daysIdsArray.length; ++i ) 
				{
					// Если текущий элемент просто одно значение - тогда добавляем его
					if ( daysIdsArray[i].length() == 1 )
					{
						result.add( daysIdsArray[i] );
					}
					// Иначе если это диапазон значений - берем границы и добавляем все значения из этого диапазона 
					else 
					{
						// начало и конец диапазона. Без проверки, т.к. диапазон должен быть вида "Значение-Значение"
						int beginId = Character.getNumericValue( daysIdsArray[i].charAt( 0 ) );
						int endId = Character.getNumericValue( daysIdsArray[i].charAt( 2 ) );
						
						for ( int j = beginId; j <= endId; ++j )
						{
							result.add( String.valueOf( j ) );
						}
					}
				}
			}
			// Если что-то вдруг пошло не так, то на странице не будут отмечены дни.
			// Это не хорошо, но не думаю, что стоит сообщать пользователю об этом
			catch ( Exception exception )
			{ }
		}
		
		return result;
	}
}
