/*
 * Copyright (c) 2016 BiTel Inc. All rights reserved.
*/
package ru.bitel.mybgbilling.plugins.documents;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.TimeUnit;
import java.util.function.Supplier;

import javax.activation.DataHandler;
import javax.enterprise.context.SessionScoped;
import javax.inject.Inject;
import javax.inject.Named;
import javax.xml.ws.Holder;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import ru.bitel.bgbilling.common.BGException;
import ru.bitel.bgbilling.common.BGIllegalAccessException;
import ru.bitel.bgbilling.kernel.directory.api.common.bean.Directory;
import ru.bitel.bgbilling.kernel.filestorage.common.bean.BGServerFile;
import ru.bitel.bgbilling.plugins.documents.common.service.DocumentsService;
import ru.bitel.bgbilling.plugins.documents.common.bean.Document;
import ru.bitel.bgbilling.plugins.documents.common.bean.DocumentJournal;
import ru.bitel.bgbilling.plugins.documents.common.bean.DocumentStatus;
import ru.bitel.bgbilling.plugins.documents.common.bean.DocumentType;
import ru.bitel.common.Utils;
import ru.bitel.common.concurrent.ConcurrentUtils;
import ru.bitel.common.function.Async;
import ru.bitel.common.function.Expirable;
import ru.bitel.common.model.Period;
import ru.bitel.mybgbilling.kernel.common.AbstractBean;
import ru.bitel.mybgbilling.kernel.common.Configuration;
import ru.bitel.mybgbilling.kernel.common.inject.BGInject;
import ru.bitel.mybgbilling.kernel.common.inject.BGInjection;

/**
 * Бин - документы договора (плагин Documents).
 * 
 * @author marsel
 *
 */
@Named
@SessionScoped
@BGInjection
public class DocumentsBean
	extends AbstractBean
	implements Serializable
{
	private static final long serialVersionUID = 1540623437095557110L;

	private static final Logger logger = LogManager.getLogger();
	
	@Inject
	private Configuration configuration;

	@BGInject
	private DocumentsService documentsService;

	@BGInject(type = DocumentJournal.class)
	private Directory<DocumentJournal> journalDirectory;

	@BGInject(type = DocumentType.class)
	private Directory<DocumentType> typeDirectory;

	/**
	 * Cписок статусов документа
	 */
	@BGInject(type = DocumentStatus.class)
	private Directory<DocumentStatus> statusDirectory;

	/**
	 * Список документов договора
	 */
	private Supplier<List<DocumentFiles>> documents = () -> Collections.emptyList();

	@Override
	protected void init()
		throws BGException
	{
		super.init();

		// XXX
		if( configuration.isPluginEnabled( "documents" ) )
		{
			populateDocuments();
		}
	}

	/**
	 * Получение списка документов плагина у договора
	 * 
	 * @return
	 * @throws BGException
	 */
	public List<DocumentFiles> getDocuments()
		throws BGException
	{
		return documents.get();
	}

	/**
	 * Получение названия статуса документа по id
	 * 
	 * @param statusId
	 * @return
	 * @throws BGException
	 */
	public String getStatusById( int statusId )
		throws BGException
	{
		DocumentStatus status = statusDirectory.get( statusId );
		if( status != null )
		{
			return status.getTitle();
		}

		return "";
	}

	public void populate()
		throws BGException
	{
	}

	public void populateDocuments()
		throws BGException
	{
		logger.info( "Получение документов по договору с id = " + getContractId() );

		documents = Expirable.asyncOf( () -> {

			// Получение документов по договору
			List<Document> documentList = Utils.maskNull( documentsService.getDocumentList( getContractId(), null, new Period() ) );
			List<DocumentFiles> documents = new ArrayList<>();

			List<CompletableFuture<?>> futureList = new ArrayList<>();

			/*
			 * Получение всей необходимой информации для вывода на странице:
			 * названия документа, его даты, тип, журнал и список прикрепленных
			 * к нему файлов, истории изменения статусов
			 */
			for( Document document : documentList )
			{
				DocumentFiles documentFiles = new DocumentFiles();
				documentFiles.setId( document.getId() );
				documentFiles.setDate( document.getCreateDate() );
				documentFiles.setTitle( document.getTitle() );

				// Вызов каждого сервиса в отдельном потоке
				futureList.add( Async.of( () -> documentsService.getDocumentFileList( document.getId() ) ).thenAccept( ( a ) -> documentFiles.setFiles( a ) ).getFuture() );
				futureList.add( Async.of( () -> documentsService.getDocumentStatusLogList( document.getId() ) ).thenAccept( ( a ) -> documentFiles.setStatusLog( a ) ).getFuture() );

				documentFiles.setType( typeDirectory.get( document.getTypeId() ) );
				documentFiles.setJournal( journalDirectory.get( document.getJournalId() ) );

				//Async<List<BGServerFile>> fileList = Async.of( () -> documentsService.getDocumentFileList( document.getId() ) ).thenAccept( (a)-> documentFiles.setFiles( a ) );
				//Async<List<DocumentStatusLogItem>> documentStatusLogItem = Async.of( () -> documentsService.getDocumentStatusLogList( document.getId() ) );
				//Async<DocumentType> documentType = Async.of( () -> documentsService.documentTypeGet( document.getTypeId() ) );
				//Async<DocumentJournal> documentJournal = Async.of( () -> documentsService.documentJournalGet( document.getJournalId() ) );

				//documentFiles.setFiles( fileList.get() );
				//documentFiles.setStatusLog( documentStatusLogItem.get() );
				//documentFiles.setType( documentType.get() );
				//documentFiles.setJournal( documentJournal.get() );

				documents.add( documentFiles );
			}

			try
			{
				ConcurrentUtils.awaitFutures( futureList );
			}
			catch( Exception e )
			{
				throw new BGException( e );
			}

			return documents;

		} , 25, TimeUnit.SECONDS );
	}

	/**
	 * Скачивание файла
	 * 
	 * @param fileId
	 *            - id файла
	 * @throws BGException
	 */
	public void downloadFile( String fileId )
		throws BGException
	{
		logger.info( "Скачивание файла с id = " + fileId + ", привязаного к договору с id = " + getContractId() );

		if( Utils.isBlankString( fileId ) )
		{
			throw new BGIllegalAccessException();
		}

		final Holder<DataHandler> holder = new Holder<DataHandler>();

		BGServerFile file = documentsService.downloadDocumentFile( Utils.parseInt( fileId ), getContractId(), holder );

		if( file == null )
		{
			throw new BGIllegalAccessException();
		}

		writeToOutput( holder.value, file.getTitle(), false );
	}
}
